# -*- coding: utf-8 -*-
from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, float_input, num_input


def isolation_forest_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Estimators: The number of trees in the forest.")
    print("Please specify the number of trees in the forest. A good starting range could be between 50 and 500, such as 100.")
    n_estimators = num_input(SECTION[2], "@N Estimators: ")
    print("Contamination: The amount of contamination of the data set.")
    print("Please specify the contamination of the data set. A good starting range could be between 0.1 and 0.5, such as 0.3.")
    contamination = float_input(0.3, SECTION[2], "@Contamination: ")
    print("Max Features: The number of features to draw from X to train each base estimator.")
    print("Please specify the number of features. A good starting range could be between 1 and the total number of features in the dataset.")
    max_features = num_input(SECTION[2], "@Max Features: ")
    print(
        "Bootstrap: Whether bootstrap samples are used when building trees. Bootstrapping is a technique where a random subset of the data is sampled with replacement"
        " to create a new dataset of the same size as the original. This new dataset is then used to construct a decision tree in the ensemble. If False, the whole dataset is used to build each tree."
    )
    print("Please specify whether bootstrap samples are used when building trees. It is generally recommended to leave it as True.")
    bootstrap = bool_input(SECTION[2])
    max_samples = None
    if bootstrap:
        print("Max Samples: The number of samples to draw from X_train to train each base estimator.")
        print("Please specify the number of samples. A good starting range could be between 256 and the number of dataset.")
        max_samples = num_input(SECTION[2], "@@Max Samples: ")
    hyper_parameters = {
        "n_estimators": n_estimators,
        "contamination": contamination,
        "max_features": max_features,
        "bootstrap": bootstrap,
    }
    if not max_samples:
        # Use the default value provided by sklearn.ensemble.RandomForestClassifier.
        hyper_parameters["max_samples"] = None
    else:
        hyper_parameters["max_samples"] = max_samples
    return hyper_parameters



# -*- coding: utf-8 -*-
from typing import Dict, List, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from rich import print

from ....constants import SECTION
from ....data.data_readiness import num_input, str_input


def pca_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Components: This parameter specifies the number of components to retain after dimensionality reduction.")
    print("Please specify the number of components to retain. A good starting range could be between 2 and 10, such as 4.")
    n_components = num_input(SECTION[2], "N Components: ")
    print("SVD Solver: This parameter specifies the algorithm used to perform the singular value decomposition.")
    print("Please specify the algorithm. It is generally recommended to leave it as 'auto'.")
    svd_solvers = ["auto", "full", "arpack", "randomized"]
    svd_solver = str_input(svd_solvers, SECTION[2])
    hyper_parameters = {"n_components": n_components, "svd_solver": svd_solver}
    return hyper_parameters


def biplot(reduced_data: pd.DataFrame, pc: pd.DataFrame, algorithm_name: str, labels: Optional[List[str]] = None) -> None:
    """Plot a compositional bi-plot for two principal components.

    Parameters
    ----------
    reduced_data : pd.DataFrame (n_samples, n_components)
        Data processed by PCA.

    pc : pd.DataFrame (n_features, n_components)
        principal components.

    algorithm_name : str
        the name of the algorithm

    labels : List[str]
        The type of tag of the samples in the data set.
    """
    plt.figure(figsize=(14, 10))

    x = reduced_data.iloc[:, 0]  # features' contributions for PC1
    y = reduced_data.iloc[:, 1]  # features' contributions for PC2
    scalex = 1.0 / (x.max() - x.min())
    scaley = 1.0 / (y.max() - y.min())

    # Draw a data point projection plot that is projected to
    # a two-dimensional plane using normal PCA
    if labels:
        legend = []
        classes = np.unique(labels)
        for i, label in enumerate(classes):
            plt.scatter(x[labels == label] * scalex, y[labels == label] * scaley, linewidth=0.01)
            legend.append("Label: {}".format(label))
        plt.legend(legend)
    else:
        plt.scatter(x * scalex, y * scaley, linewidths=0.01)

    # principal component's weight coefficient
    # it can obtain the expression of principal component in feature space
    n = pc.shape[0]

    # plot arrows as the variable contribution,
    # each variable has a score for PC1 and for PC2 respectively
    for i in range(n):
        plt.arrow(
            0,
            0,
            pc.iloc[i, 0],
            pc.iloc[i, 1],
            color="k",
            alpha=0.7,
            linewidth=1,
        )
        plt.text(pc.iloc[i, 0] * 1.01, pc.iloc[i, 1] * 1.01, pc.index[i], ha="center", va="center", color="k", fontsize=12)

    plt.xlabel(f"{pc.columns[0]}")
    plt.ylabel(f"{pc.columns[1]}")
    plt.title(f"Compositional Bi-plot - {algorithm_name}")
    plt.grid()


def triplot(reduced_data: pd.DataFrame, pc: pd.DataFrame, algorithm_name: str, labels: Optional[List[str]] = None) -> None:
    """Plot a compositional tri-plot in 3d for three principal components.

    Parameters
    ----------
    reduced_data : pd.DataFrame (n_samples, n_components)
        Data processed by PCA.

    pc : pd.DataFrame (n_features, n_components)
        principal components.

    algorithm_name : str
        the name of the algorithm

    store_path : str
        the local path to store the graph produced

    labels : List[str]
        The type of tag of the samples in the data set.
    """

    plt.figure(figsize=(14, 12))
    ax = plt.axes(projection="3d")

    x = reduced_data.iloc[:, 0]  # features' contributions for PC1
    y = reduced_data.iloc[:, 1]  # features' contributions for PC2
    z = reduced_data.iloc[:, 2]  # features' contributions for PC3
    scalex = 1.0 / (x.max() - x.min())
    scaley = 1.0 / (y.max() - y.min())
    scalez = 1.0 / (z.max() - z.min())

    # Draw a data point projection plot that is projected to
    # a three-dimensional space using normal PCA
    if labels:
        legend = []
        classes = np.unique(labels)  # label type
        for i, label in enumerate(classes):
            ax.scatter3D(x[labels == label] * scalex, y[labels == label] * scaley, z[labels == label] * scalez, linewidth=0.01)
            # hyperparameter in plt.scatter(): c=colors[i], marker=markers[i]
            legend.append("Label: {}".format(label))
        ax.legend(legend)
    else:
        ax.scatter3D(x * scalex, y * scaley, z * scalez, linewidths=0.01)

    # the initial angle to draw the 3d plot
    azim = -60  # azimuth
    elev = 30  # elevation
    ax.view_init(elev, azim)  # set the angles

    # principal component's weight coefficient
    # it can obtain the expression of principal component in feature space
    n = pc.shape[0]

    # plot arrows as the column_name contribution,
    # each column_name has a score for PC1, for PC2 and for PC3 respectively
    for i in range(n):
        ax.quiver(0, 0, 0, pc.iloc[i, 0], pc.iloc[i, 1], pc.iloc[i, 2], color="k", alpha=0.7, linewidth=1, arrow_length_ratio=0.05)
        ax.text(
            pc.iloc[i, 0] * 1.1,
            pc.iloc[i, 1] * 1.1,
            pc.iloc[i, 2] * 1.1,
            pc.index[i],
            ha="center",
            va="center",
            color="k",
            fontsize=12,
        )

    ax.set_xlabel(f"{pc.columns[0]}")
    ax.set_ylabel(f"{pc.columns[1]}")
    ax.set_zlabel(f"{pc.columns[2]}")
    plt.title(f"Compositional Tri-plot - {algorithm_name}")
    plt.grid()



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, num_input


def mds_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Components: This parameter specifies the number of components to retain after dimensionality reduction.")
    print("Please specify the number of components to retain. A good starting range could be between 2 and 10, such as 4.")
    n_components = num_input(SECTION[2], "N Components: ")
    print("Metric: This parameter specifies the metric to be used when calculating distance between instances in a feature array.")
    print("Please specify whether the metric is used when measuring the pairwise distances between data points in the input space. It is generally recommended to leave it as True.")
    metric = bool_input(SECTION[2])
    print("N Init: This parameter specifies the number of times the SMACOF algorithm will be run with different initializations.")
    print("Please specify the number of times. A good starting range could be between 1 and 10, such as 4.")
    n_init = num_input(SECTION[2], "N Init: ")
    print("Max Iter: This parameter specifies the maximum number of iterations of the SMACOF algorithm for a single run.")
    print("Please specify the maximum number of iterations. A good starting range could be between 100 and 1000, such as 300.")
    max_iter = num_input(SECTION[2], "Max Iter: ")
    hyper_parameters = {"n_components": n_components, "metric": metric, "n_init": n_init, "max_iter": max_iter}
    return hyper_parameters



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from rich import print
from sklearn.impute import SimpleImputer

from ..constants import SECTION
from .data_readiness import float_input


def imputer(data: pd.DataFrame, method: str) -> tuple[dict, np.ndarray]:
    """Apply imputation on missing values.

    Parameters
    ----------
    data : pd.DataFrame
        The dataset with missing values.

    method : str
        The imputation method.

    Returns
    -------
    imputation_config : dict
        The imputation configuration.

    data_imputed : np.ndarray
        The dataset after imputing.
    """
    method2option = {
        "Mean Value": "mean",
        "Median Value": "median",
        "Most Frequent Value": "most_frequent",
        "Constant(Specified Value)": "constant",
    }[method]
    if method2option == "constant":
        filled_value = float_input(0, SECTION[2], "@Specified Value: ")
        imputer = SimpleImputer(missing_values=np.nan, strategy=method2option, fill_value=filled_value)
    else:
        imputer = SimpleImputer(missing_values=np.nan, strategy=method2option)
    print(f"Successfully fill the missing values with the {method2option} value " f"of each feature column respectively.")
    data_imputed = imputer.fit_transform(data)
    imputation_config = {type(imputer).__name__: imputer.get_params()}
    return imputation_config, data_imputed



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input, str_input


def dbscan_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("Eps: The maximum distance between two samples for one to be considered as in the neighborhood of the other.")
    print("Please specify the maximum distance. A good starting range could be between 0.1 and 1.0, such as 0.5.")
    eps = float_input(0.5, SECTION[2], "Eps: ")

    print("Min Samples: The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.")
    print("Please specify the number of samples. A good starting range could be between 5 and 20, such as 5.")
    min_samples = num_input(SECTION[2], "Min Samples: ")

    print("Algorithm: The algorithm to be used by the NearestNeighbors module to compute pointwise distances and find nearest neighbors.")
    print("Please specify the algorithm. It is generally recommended to leave it as 'auto'.")
    algorithms = ["auto", "ball_tree", "kd_tree", "brute"]
    algorithm = str_input(algorithms, SECTION[2])

    # Reference:
    # https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    # https://scikit-learn.org/stable/modules/neighbors.html
    print("Metric: The metric to use when calculating distance between instances in a feature array.")
    print("Please specify the metric to use when calculating distance between instances in a feature array. It is generally recommended to leave it as 'euclidean'.")
    if algorithm == "kd_tree":
        metrics = ["euclidean", "l2", "minkowski", "p", "manhattan", "cityblock", "l1", "chebyshev", "infinity"]
    elif algorithm == "ball_tree":
        metrics = [
            "euclidean",
            "l2",
            "minkowski",
            "p",
            "manhattan",
            "cityblock",
            "l1",
            "chebyshev",
            "infinity",
            "seuclidean",
            "mahalanobis",
            "hamming",
            "canberra",
            "braycurtis",
            "jaccard",
            "dice",
            "rogerstanimoto",
            "russellrao",
            "sokalmichener",
            "sokalsneath",
            "haversine",
        ]
    else:
        metrics = ["euclidean", "manhattan", "chebyshev", "minkowski", "cosine", "correlation"]
    metric = str_input(metrics, SECTION[2])

    print("Leaf Size: Leaf size passed to BallTree or KDTree. This can affect the speed of the construction and query, as well as the memory required to store the tree.")
    print("Please specify the leaf size. A good starting range could be between 10 and 30, such as 30.")
    leaf_size = num_input(SECTION[2], "Leaf Size: ")
    p = None
    if metric == "minkowski":
        print("P: The power of the Minkowski metric to be used to calculate distance between points.")
        print("Please specify the power of the Minkowski metric. A good starting range could be between 1 and 2, such as 2.")
        p = num_input(SECTION[2], "P: ")
    hyper_parameters = {
        "eps": eps,
        "min_samples": min_samples,
        "metric": metric,
        "algorithm": algorithm,
        "leaf_size": leaf_size,
        "p": p,
    }
    return hyper_parameters



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import num_input, str_input


def knn_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Neighbors: The number of neighbors to use by default when making predictions.")
    print(
        "Please specify the number of neighbors to use by default for kneighbors queries. A good starting value could be between 1 and"
        " the square root of the number of samples in your dataset, such as 5."
    )
    n_neighbors = num_input(SECTION[2], "@N Neighbors: ")
    print("Weights: weight function used in prediction.")
    print("Please specify the weight function used in prediction. It is generally recommended to leave it as 'uniform'.")
    weights = ["uniform", "distance"]
    weight = str_input(weights, SECTION[2])
    print("Algorithm: Algorithm used to compute the nearest neighbors.")
    print("Please specify the algorithm used to compute the nearest neighbors. It is generally recommended to leave it as 'auto'.")
    algorithms = ["auto", "ball_tree", "kd_tree", "brute"]
    algorithm = str_input(algorithms, SECTION[2])
    leaf_size = None
    if algorithm in ["ball_tree", "kd_tree"]:
        print("Leaf Size: Leaf size controls the size of leaf nodes in the tree-based algorithms ('ball_tree' and 'kd_tree')")
        print("Please specify the leaf size passed to BallTree or KDTree. A good starting value could be between 10 and 100, such as 30.")
        leaf_size = num_input(SECTION[2], "@Leaf Size: ")
    print("Metric: The distance metric to use for distance computation.")
    print("Please specify the distance metric to use for the tree. It is generally recommended to leave it as 'minkowski'.")
    metrics = ["euclidean", "manhattan", "minkowski"]
    metric = str_input(metrics, SECTION[2])
    p = None
    if metric == "minkowski":
        print("P: Power parameter for the Minkowski metric.")
        print("Please specify the power parameter for the Minkowski metric. A good starting value could be between 1 and 5, such as 2.")
        p = num_input(SECTION[2], "@P: ")
    hyper_parameters = {
        "n_neighbors": n_neighbors,
        "weights": weight,
        "algorithm": algorithm,
        "metric": metric,
    }
    if not leaf_size:
        # Use the default value provided by sklearn.neighbors.KNeighborsRegressor
        hyper_parameters["leaf_size"] = 30
    else:
        hyper_parameters["leaf_size"] = leaf_size
    if not p:
        # Use the default value provided by sklearn.neighbors.KNeighborsRegressor
        hyper_parameters["p"] = 2
    else:
        hyper_parameters["p"] = p
    return hyper_parameters



# -*- coding: utf-8 -*-
import os

import pandas as pd

from ..constants import MLFLOW_ARTIFACT_DATA_PATH
from ..model.detection import AbnormalDetectionWorkflowBase, IsolationForestAbnormalDetection
from ._base import ModelSelectionBase


class AbnormalDetectionModelSelection(ModelSelectionBase):
    """Simulate the normal way of invoking scikit-learn abnormal detection algorithms."""

    def __init__(self, model_name: str) -> None:
        self.model_name = model_name
        self.ad_workflow = AbnormalDetectionWorkflowBase()
        self.transformer_config = {}

    # @dispatch(object, object, object, object, object, object)
    def activate(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        y_train: pd.DataFrame,
        y_test: pd.DataFrame,
    ) -> None:
        """Train by Scikit-learn framework."""

        self.ad_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test)

        # Model option
        if self.model_name == "Isolation Forest":
            hyper_parameters = IsolationForestAbnormalDetection.manual_hyper_parameters()
            self.ad_workflow = IsolationForestAbnormalDetection(
                n_estimators=hyper_parameters["n_estimators"],
                contamination=hyper_parameters["contamination"],
                max_features=hyper_parameters["max_features"],
                bootstrap=hyper_parameters["bootstrap"],
                max_samples=hyper_parameters["max_samples"],
            )

        self.ad_workflow.show_info()

        # Use Scikit-learn style API to process input data
        self.ad_workflow.fit(X)
        y_predict = self.ad_workflow.predict(X)
        X_abnormal_detection, X_normal, X_abnormal = self.ad_workflow._detect_data(X, y_predict)

        self.ad_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test)

        # Save the model hyper-parameters
        self.ad_workflow.save_hyper_parameters(hyper_parameters, self.model_name, os.getenv("GEOPI_OUTPUT_PARAMETERS_PATH"))

        # Common components for every abnormal detection algorithm
        self.ad_workflow.common_components()

        # special components of different algorithms
        self.ad_workflow.special_components()

        # Save abnormal detection result
        self.ad_workflow.data_save(X_abnormal_detection, "X Abnormal Detection", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Abnormal Detection Data")
        self.ad_workflow.data_save(X_normal, "X Normal", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Normal Data")
        self.ad_workflow.data_save(X_abnormal, "X Abnormal", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Abnormal Data")

        # Save the trained model
        self.ad_workflow.model_save()



# -*- coding: utf-8 -*-
from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, float_input, num_input, str_input


def svc_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("Kernel: This hyperparameter specifies the kernel function to be used for mapping the input data to a higher-dimensional feature space.")
    print("Please specify the kernel type to be used in the algorithm. It is generally recommended to leave it as 'Radial Basis Function (RBF) Kernel'.")
    kernels = ["linear", "poly", "rbf", "sigmoid"]
    kernel = str_input(kernels, SECTION[2])
    degree = None
    if kernel == "poly":
        print("Degree: This hyperparameter is only used when the kernel is set to polynomial. It specifies the degree of the polynomial kernel function.")
        print("Please specify the degree of the polynomial kernel function. A good starting range could be between 2 and 5, such as 3.")
        degree = num_input(SECTION[2], "@Degree: ")
    gamma = None
    if kernel in ["poly", "rbf", "sigmoid"]:
        print("Gamma: This hyperparameter is only used when the kernel is set to Polynomial, RBF, or Sigmoid. It specifies the kernel coefficient for rbf, poly and sigmoid.")
        print("Please specify the kernel coefficient for rbf, poly and sigmoid. A good starting range could be between 0.0001 and 10, such as 0.1.")
        gamma = float_input(0.1, SECTION[2], "@Gamma: ")
    # coef0 = None
    # if kernel in ["poly", "sigmoid"]:
    #     print("Coef0: This hyperparameter is only used when the kernel is set to Polynomial or Sigmoid. It specifies the independent term in kernel function.")
    #     print("The coef0 parameter controls the influence of higher-order terms in the polynomial and sigmoid kernels, and it can help to adjust the balance between the linear and nonlinear parts"
    #       " of the decision boundary.")
    #     print("Please specify the independent term in kernel function. A good starting range could be between 0.0 and 1.0, such as 0.0.")
    #     coef0 = float_input(SECTION[2], "@Coef0: ")
    print("C: This hyperparameter specifies the penalty parameter C of the error term.")
    print("The C parameter controls the trade off between smooth decision boundary and classifying the training points correctly.")
    print("Please specify the penalty parameter C of the error term. A good starting range could be between 0.001 and 1000, such as 1.0.")
    C = float_input(1, SECTION[2], "@C: ")
    print("Shrinking: This hyperparameter specifies whether to use the shrinking heuristic.")
    print("The shrinking heuristic is a technique that speeds up the training process by only considering the support vectors in the decision function.")
    print("Please specify whether to use the shrinking heuristic. It is generally recommended to leave it as True.")
    shrinking = bool_input(SECTION[2])
    hyper_parameters = {"kernel": kernel, "C": C, "shrinking": shrinking}
    if not degree:
        # Use the default value provided by sklearn.svm.SVC
        hyper_parameters["degree"] = 3
    else:
        hyper_parameters["degree"] = degree
    if not gamma:
        # Use the default value provided by sklearn.svm.SVC
        hyper_parameters["gamma"] = "scale"
    else:
        hyper_parameters["gamma"] = gamma
    # if coef0:
    #     # Use the default value provided by sklearn.svm.SVC
    #     hyper_parameters["coef0"] = 0.0
    return hyper_parameters



from database import get_db
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from sqlalchemy.orm import Session

from .schemas import User
from .service import get_user_by_email
from .utils import decode_access_token

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/auth/login", scheme_name="JWT")


async def get_current_user(token: str = Depends(oauth2_scheme), db: Session = Depends(get_db)):
    token_data = decode_access_token(token)
    # user = get_user_by_email(db, email=token_data.email)
    user = get_user_by_email(db, email=token_data)
    if user is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Could not validate credentials", headers={"WWW-Authenticate": "Bearer"})
    return user


async def get_current_active_user(current_user: User = Depends(get_current_user)):
    if not current_user.is_active:
        raise HTTPException(status_code=400, detail="Inactive user")
    return current_user



# -*- coding: utf-8 -*-
import json
import os
from typing import Dict, Optional, Union

import mlflow
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from rich import print
from sklearn.cluster import DBSCAN, AffinityPropagation, AgglomerativeClustering, KMeans

from ..constants import MLFLOW_ARTIFACT_DATA_PATH, MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH
from ..utils.base import clear_output, save_data, save_fig, save_text
from ._base import WorkflowBase
from .func.algo_clustering._affinitypropagation import affinitypropagation_manual_hyper_parameters
from .func.algo_clustering._agglomerative import agglomerative_manual_hyper_parameters
from .func.algo_clustering._common import plot_silhouette_diagram, plot_silhouette_value_diagram, scatter2d, scatter3d, score
from .func.algo_clustering._dbscan import dbscan_manual_hyper_parameters
from .func.algo_clustering._kmeans import kmeans_manual_hyper_parameters


class ClusteringWorkflowBase(WorkflowBase):
    """The base workflow class of clustering algorithms."""

    common_function = ["Cluster Centers", "Cluster Labels", "Model Persistence"]

    def __init__(self):
        super().__init__()
        self.clustering_result = None
        self.mode = "Clustering"

    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> None:
        """Fit the model according to the given training data."""
        self.X = X
        self.model.fit(X)
        mlflow.log_params(self.model.get_params())

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        return dict()

    # TODO(Samson 1057266013@qq.com): This function might need to be rethought.
    def get_cluster_centers(self) -> np.ndarray:
        """Get the cluster centers."""
        print("-----* Clustering Centers *-----")
        print(getattr(self.model, "cluster_centers_", "This class don not have cluster_centers_"))
        return getattr(self.model, "cluster_centers_", "This class don not have cluster_centers_")

    def get_labels(self):
        """Get the cluster labels."""
        print("-----* Clustering Labels *-----")
        # self.X['clustering result'] = self.model.labels_
        self.clustering_result = pd.DataFrame(self.model.labels_, columns=["clustering result"])
        print(self.clustering_result)
        GEOPI_OUTPUT_ARTIFACTS_DATA_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH")
        save_data(self.clustering_result, f"{self.naming} Result", GEOPI_OUTPUT_ARTIFACTS_DATA_PATH, MLFLOW_ARTIFACT_DATA_PATH)

    @staticmethod
    def _score(data: pd.DataFrame, labels: pd.DataFrame, algorithm_name: str, store_path: str) -> None:
        """Calculate the score of the model."""
        print("-----* Model Score *-----")
        scores = score(data, labels)
        scores_str = json.dumps(scores, indent=4)
        save_text(scores_str, f"Model Score - {algorithm_name}", store_path)
        mlflow.log_metrics(scores)

    @staticmethod
    def _scatter2d(data: pd.DataFrame, labels: pd.DataFrame, cluster_centers_: np.ndarray, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the two-dimensional diagram of the clustering result."""
        print("-----* Cluster Two-Dimensional Diagram *-----")
        scatter2d(data, labels, cluster_centers_, algorithm_name)
        save_fig(f"Cluster Two-Dimensional Diagram - {algorithm_name}", local_path, mlflow_path)
        data_with_labels = pd.concat([data, labels], axis=1)
        save_data(data_with_labels, f"Cluster Two-Dimensional Diagram - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _scatter3d(data: pd.DataFrame, labels: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the three-dimensional diagram of the clustering result."""
        print("-----* Cluster Three-Dimensional Diagram *-----")
        scatter3d(data, labels, algorithm_name)
        save_fig(f"Cluster Three-Dimensional Diagram - {algorithm_name}", local_path, mlflow_path)
        data_with_labels = pd.concat([data, labels], axis=1)
        save_data(data_with_labels, f"Cluster Two-Dimensional Diagram - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _plot_silhouette_diagram(data: pd.DataFrame, labels: pd.DataFrame, model: object, cluster_centers_: np.ndarray, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the silhouette diagram of the clustering result."""
        print("-----* Silhouette Diagram *-----")
        plot_silhouette_diagram(data, labels, cluster_centers_, model, algorithm_name)
        save_fig(f"Silhouette Diagram - {algorithm_name}", local_path, mlflow_path)
        data_with_labels = pd.concat([data, labels], axis=1)
        save_data(data_with_labels, "Silhouette Diagram - Data With Labels", local_path, mlflow_path)
        if not isinstance(cluster_centers_, str):
            cluster_center_data = pd.DataFrame(cluster_centers_, columns=data.columns)
            save_data(cluster_center_data, "Silhouette Diagram - Cluster Centers", local_path, mlflow_path)

    @staticmethod
    def _plot_silhouette_value_diagram(data: pd.DataFrame, labels: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the silhouette value diagram of the clustering result."""
        print("-----* Silhouette value Diagram *-----")
        plot_silhouette_value_diagram(data, labels, algorithm_name)
        save_fig(f"Silhouette value Diagram - {algorithm_name}", local_path, mlflow_path)
        data_with_labels = pd.concat([data, labels], axis=1)
        save_data(data_with_labels, "Silhouette value Diagram - Data With Labels", local_path, mlflow_path)

    def common_components(self) -> None:
        """Invoke all common application functions for clustering algorithms."""
        GEOPI_OUTPUT_METRICS_PATH = os.getenv("GEOPI_OUTPUT_METRICS_PATH")
        GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH")
        self._score(
            data=self.X,
            labels=self.clustering_result["clustering result"],
            algorithm_name=self.naming,
            store_path=GEOPI_OUTPUT_METRICS_PATH,
        )
        if self.X.shape[1] >= 3:
            # choose two of dimensions to draw
            two_dimen_axis_index, two_dimen_data = self.choose_dimension_data(self.X, 2)
            self._scatter2d(
                data=two_dimen_data,
                labels=self.clustering_result["clustering result"],
                cluster_centers_=self.get_cluster_centers(),
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )

            # choose three of dimensions to draw
            three_dimen_axis_index, three_dimen_data = self.choose_dimension_data(self.X, 3)
            self._scatter3d(
                data=three_dimen_data,
                labels=self.clustering_result["clustering result"],
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
        elif self.X.shape[1] == 3:
            # choose two of dimensions to draw
            two_dimen_axis_index, two_dimen_data = self.choose_dimension_data(self.X, 2)
            self._scatter2d(
                data=two_dimen_data,
                labels=self.clustering_result["clustering result"],
                cluster_centers_=self.get_cluster_centers(),
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )

            # no need to choose
            self._scatter3d(
                data=self.X,
                labels=self.clustering_result["clustering result"],
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
        elif self.X.shape[1] == 2:
            self._scatter2d(
                data=self.X,
                labels=self.clustering_result["clustering result"],
                cluster_centers_=self.get_cluster_centers(),
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
        else:
            pass

        self._plot_silhouette_diagram(
            data=self.X,
            labels=self.clustering_result["clustering result"],
            cluster_centers_=self.get_cluster_centers(),
            model=self.model,
            algorithm_name=self.naming,
            local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
            mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
        )
        self._plot_silhouette_value_diagram(
            data=self.X,
            labels=self.clustering_result["clustering result"],
            algorithm_name=self.naming,
            local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
            mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
        )


class KMeansClustering(ClusteringWorkflowBase):
    """The automation workflow of using KMeans algorithm to make insightful products."""

    name = "KMeans"
    special_function = ["KMeans Score"]

    def __init__(
        self,
        n_clusters: int = 8,
        init: str = "k-means++",
        n_init: int = 10,
        max_iter: int = 300,
        tol: float = 1e-4,
        verbose: int = 0,
        random_state: Optional[int] = None,
        copy_x: bool = True,
        algorithm: str = "auto",
    ) -> None:
        """
        Parameters
        ----------
        n_clusters : int, default=8
            The number of clusters to form as well as the number of
            centroids to generate.

        init : {'k-means++', 'random'}, callable or array-like of shape \
                (n_clusters, n_features), default='k-means++'
            Method for initialization:

            'k-means++' : selects initial cluster centers for k-mean
            clustering in a smart way to speed up convergence. See section
            Notes in k_init for more details.

            'random': choose `n_clusters` observations (rows) at random from data
            for the initial centroids.

            If an array is passed, it should be of shape (n_clusters, n_features)
            and gives the initial centers.

            If a callable is passed, it should take arguments X, n_clusters and a
            random state and return an initialization.

        n_init : int, default=10
            Number of time the k-means algorithm will be run with different
            centroid seeds. The final results will be the best output of
            n_init consecutive runs in terms of inertia.

        max_iter : int, default=300
            Maximum number of iterations of the k-means algorithm for a
            single run.

        tol : float, default=1e-4
            Relative tolerance with regards to Frobenius norm of the difference
            in the cluster centers of two consecutive iterations to declare
            convergence.

        verbose : int, default=0
            Verbosity mode.

        random_state : int, RandomState instance or None, default=None
            Determines random number generation for centroid initialization. Use
            an int to make the randomness deterministic.
            See :term:`Glossary <random_state>`.

        copy_x : bool, default=True
            When pre-computing distances it is more numerically accurate to center
            the data first. If copy_x is True (default), then the original data is
            not modified. If False, the original data is modified, and put back
            before the function returns, but small numerical differences may be
            introduced by subtracting and then adding the data mean. Note that if
            the original data is not C-contiguous, a copy will be made even if
            copy_x is False. If the original data is sparse, but not in CSR format,
            a copy will be made even if copy_x is False.

        algorithm : {"auto", "full", "elkan"}, default="auto"
            K-means algorithm to use. The classical EM-style algorithm is "full".
            The "elkan" variation is more efficient on data with well-defined
            clusters, by using the triangle inequality. However it's more memory
            intensive due to the allocation of an extra array of shape
            (n_samples, n_clusters).

            For now "auto" (kept for backward compatibility) chooses "elkan" but it
            might change in the future for a better heuristic.

        References
        ----------------------------------------
        Scikit-learn API: sklearn.cluster.KMeans
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
        """

        super().__init__()
        self.n_clusters = n_clusters
        self.init = init
        self.max_iter = max_iter
        self.tol = tol
        self.n_init = n_init
        self.verbose = verbose
        self.copy_x = copy_x
        self.algorithm = algorithm

        if random_state:
            self.random_state = random_state

        # If 'random_state' is None, 'self.random_state' comes from the parent class 'WorkflowBase'
        self.model = KMeans(
            n_clusters=self.n_clusters,
            init=self.init,
            n_init=self.n_init,
            max_iter=self.max_iter,
            tol=self.tol,
            verbose=self.verbose,
            random_state=self.random_state,
            copy_x=self.copy_x,
            algorithm=self.algorithm,
        )

        self.naming = KMeansClustering.name

    def _get_inertia_scores(self, algorithm_name: str, store_path: str) -> None:
        """Get the scores of the clustering result."""
        print("-----* KMeans Inertia Scores *-----")
        print("Inertia Score: ", self.model.inertia_)
        inertia_scores = {"Inertia Score": self.model.inertia_}
        mlflow.log_metrics(inertia_scores)
        inertia_scores_str = json.dumps(inertia_scores, indent=4)
        save_text(inertia_scores_str, f"KMeans Inertia Scores - {algorithm_name}", store_path)

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = kmeans_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs: Union[Dict, np.ndarray, int]) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""
        GEOPI_OUTPUT_METRICS_PATH = os.getenv("GEOPI_OUTPUT_METRICS_PATH")
        self._get_inertia_scores(
            algorithm_name=self.naming,
            store_path=GEOPI_OUTPUT_METRICS_PATH,
        )


class DBSCANClustering(ClusteringWorkflowBase):
    """The automation workflow of using DBSCAN algorithm to make insightful products."""

    name = "DBSCAN"
    special_function = ["Virtualization of Result in 2D Graph"]

    def __init__(
        self,
        eps: float = 0.5,
        min_samples: int = 5,
        metric: str = "euclidean",
        metric_params: Optional[Dict] = None,
        algorithm: str = "auto",
        leaf_size: int = 30,
        p: float = None,
        n_jobs: int = None,
    ) -> None:
        """
        Parameters
        ----------
        eps : float, default=0.5
            The maximum distance between two samples for one to be considered as in the neighborhood of the other. This is not a maximum bound on the distances of points within a cluster.
            This is the most important DBSCAN parameter to choose appropriately for your data set and distance function.

        min_samples : int, default=5
            The number of samples (or total weight) in a neighborhood for a point to be considered as a core point. This includes the point itself.

        metric : str, or callable, default=`euclidean`
            The metric to use when calculating distance between instances in a feature array. If metric is a string or callable, it must be one of the options allowed
              by sklearn.metrics.pairwise_distances for its metric parameter.
            If metric is “precomputed”, X is assumed to be a distance matrix and must be square. X may be a sparse graph, in which case only “nonzero” elements may be considered neighbors for DBSCAN.

            New in version 0.17: metric precomputed to accept precomputed sparse matrix.

        metric_params : dict, default=None
            Additional keyword arguments for the metric function.

            New in version 0.19.

        algorithm : {`auto`, `ball_tree`, `kd_tree`, `brute`}, default=`auto`
            The algorithm to be used by the NearestNeighbors module to compute pointwise distances and find nearest neighbors. See NearestNeighbors module documentation for details.

        leaf_size : int, default=30
            Leaf size passed to BallTree or cKDTree. This can affect the speed of the construction and query, as well as the memory required to store the tree. The optimal value depends
            on the nature of the problem.

        p : float, default=None
            The power of the Minkowski metric to be used to calculate distance between points. If None, then p=2 (equivalent to the Euclidean distance).

        n_jobs : int, default=None
            The number of parallel jobs to run. None means 1 unless in a joblib.parallel_backend context. -1 means using all processors. See Glossary for more details.

        References
        ----------------------------------------
        Scikit-learn API: sklearn.cluster.DBSCAN
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
        """

        super().__init__()
        self.eps = eps
        self.min_samples = min_samples
        self.metric = metric
        self.metric_params = metric_params
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.p = p
        self.n_jobs = n_jobs

        self.model = DBSCAN(
            eps=self.eps,
            min_samples=self.min_samples,
            metric=self.metric,
            metric_params=self.metric_params,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            p=self.p,
            n_jobs=self.n_jobs,
        )

        self.naming = DBSCANClustering.name

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = dbscan_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs: Union[Dict, np.ndarray, int]) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""


class Agglomerative(ClusteringWorkflowBase):
    """The automation workflow of using Agglomerative Clustering to make insightful products."""

    name = "Agglomerative"
    special_function = []

    def __init__(
        self,
        n_clusters: int = 2,
        *,
        affinity: str = "euclidean",
        memory: str = None,
        connectivity: ArrayLike = None,
        compute_full_tree: str = "auto",
        linkage: str = "ward",
        distance_threshold: float = None,
        compute_distances: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        n_clusters : int or None, default=2
            The number of clusters to find. It must be ``None`` if
            ``distance_threshold`` is not ``None``.

        affinity : str or callable, default='euclidean'
            Metric used to compute the linkage. Can be "euclidean", "l1", "l2",
            "manhattan", "cosine", or "precomputed".
            If linkage is "ward", only "euclidean" is accepted.
            If "precomputed", a distance matrix (instead of a similarity matrix)
            is needed as input for the fit method.

        memory : str or object with the joblib.Memory interface, default=None
            Used to cache the output of the computation of the tree.
            By default, no caching is done. If a string is given, it is the
            path to the caching directory.

        connectivity : array-like or callable, default=None
            Connectivity matrix. Defines for each sample the neighboring
            samples following a given structure of the data.
            This can be a connectivity matrix itself or a callable that transforms
            the data into a connectivity matrix, such as derived from
            `kneighbors_graph`. Default is ``None``, i.e, the
            hierarchical clustering algorithm is unstructured.

        compute_full_tree : 'auto' or bool, default='auto'
            Stop early the construction of the tree at ``n_clusters``. This is
            useful to decrease computation time if the number of clusters is not
            small compared to the number of samples. This option is useful only
            when specifying a connectivity matrix. Note also that when varying the
            number of clusters and using caching, it may be advantageous to compute
            the full tree. It must be ``True`` if ``distance_threshold`` is not
            ``None``. By default `compute_full_tree` is "auto", which is equivalent
            to `True` when `distance_threshold` is not `None` or that `n_clusters`
            is inferior to the maximum between 100 or `0.02 * n_samples`.
            Otherwise, "auto" is equivalent to `False`.

        linkage : {'ward', 'complete', 'average', 'single'}, default='ward'
            Which linkage criterion to use. The linkage criterion determines which
            distance to use between sets of observation. The algorithm will merge
            the pairs of cluster that minimize this criterion.

            - 'ward' minimizes the variance of the clusters being merged.
            - 'average' uses the average of the distances of each observation of
            the two sets.
            - 'complete' or 'maximum' linkage uses the maximum distances between
            all observations of the two sets.
            - 'single' uses the minimum of the distances between all observations
            of the two sets.

            .. versionadded:: 0.20
                Added the 'single' option

        distance_threshold : float, default=None
            The linkage distance threshold above which, clusters will not be
            merged. If not ``None``, ``n_clusters`` must be ``None`` and
            ``compute_full_tree`` must be ``True``.

            .. versionadded:: 0.21

        compute_distances : bool, default=False
            Computes distances between clusters even if `distance_threshold` is not
            used. This can be used to make dendrogram visualization, but introduces
            a computational and memory overhead.

            .. versionadded:: 0.24

        References
        ----------
        sklearn.cluster.AgglomerativeClustering
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html
        """

        super().__init__()
        self.n_clusters = n_clusters
        self.affinity = affinity
        self.distance_threshold = distance_threshold
        self.memory = memory
        self.connectivity = connectivity
        self.compute_full_tree = compute_full_tree
        self.linkage = linkage
        self.compute_distances = compute_distances

        self.model = AgglomerativeClustering(
            n_clusters=self.n_clusters,
            affinity=self.affinity,
            memory=self.memory,
            connectivity=self.connectivity,
            compute_full_tree=self.compute_full_tree,
            linkage=self.linkage,
            distance_threshold=self.distance_threshold,
            compute_distances=self.compute_distances,
        )

        self.naming = Agglomerative.name

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = agglomerative_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""
        pass


class AffinityPropagationClustering(ClusteringWorkflowBase):
    name = "AffinityPropagation"

    def __init__(
        self,
        *,
        damping: float = 0.5,
        max_iter: int = 200,
        convergence_iter: int = 15,
        copy: bool = True,
        preference: Optional[Dict] = None,
        affinity: str = "euclidean",
        verbose: bool = False,
        random_state: Optional[Dict] = None,
    ) -> None:
        """
        Parameters
        ----------
        damping : float, default=0.5
            Damping factor in the range `[0.5, 1.0)` is the extent to
            which the current value is maintained relative to
            incoming values (weighted 1 - damping). This in order
            to avoid numerical oscillations when updating these
            values (messages).

        max_iter : int, default=200
            Maximum number of iterations.

        convergence_iter : int, default=15
            Number of iterations with no change in the number
            of estimated clusters that stops the convergence.

        copy : bool, default=True
            Make a copy of input data.

        preference : array-like of shape (n_samples,) or float, default=None
            Preferences for each point - points with larger values of
            preferences are more likely to be chosen as exemplars. The number
            of exemplars, ie of clusters, is influenced by the input
            preferences value. If the preferences are not passed as arguments,
            they will be set to the median of the input similarities.

        affinity : {'euclidean', 'precomputed'}, default='euclidean'
            Which affinity to use. At the moment 'precomputed' and
            ``euclidean`` are supported. 'euclidean' uses the
            negative squared euclidean distance between points.

        verbose : bool, default=False
            Whether to be verbose.

        random_state : int, RandomState instance or None, default=None
            Pseudo-random number generator to control the starting state.
            Use an int for reproducible results across function calls.
            See the :term:`Glossary <random_state>`.

            .. versionadded:: 0.23
                this parameter was previously hardcoded as 0.

        References
        ----------------------------------------
        Scikit-learn API: sklearn.cluster.AffinityPropagation
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AffinityPropagation
        """

        super().__init__()
        self.damping = damping
        self.max_iter = max_iter
        self.convergence_iter = convergence_iter
        self.copy = copy
        self.verbose = verbose
        self.preference = preference
        self.affinity = affinity

        if random_state:
            self.random_state = random_state

        # If 'random_state' is None, 'self.random_state' comes from the parent class 'WorkflowBase'
        self.model = AffinityPropagation(
            damping=self.damping,
            max_iter=self.max_iter,
            convergence_iter=self.convergence_iter,
            copy=self.copy,
            preference=self.preference,
            affinity=self.affinity,
            verbose=self.verbose,
            random_state=self.random_state,
        )
        self.naming = AffinityPropagationClustering.name

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = affinitypropagation_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs: Union[Dict, np.ndarray, int]) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""


class MeanShiftClustering(ClusteringWorkflowBase):
    name = "MeanShift"
    pass


class SpectralClustering(ClusteringWorkflowBase):
    name = "Spectral"
    pass


class WardHierarchicalClustering(ClusteringWorkflowBase):
    name = "WardHierarchical"
    pass


class OPTICSClustering(ClusteringWorkflowBase):
    name = "OPTICS"
    pass


class GaussianMixturesClustering(ClusteringWorkflowBase):
    name = "GaussianMixtures"
    pass


class BIRCHClusteringClustering(ClusteringWorkflowBase):
    name = "BIRCHClustering"
    pass


class BisectingKMeansClustering(ClusteringWorkflowBase):
    name = "BisectingKMeans"
    pass



# -*- coding: utf-8 -*-
import os
from typing import Optional

import pandas as pd

from ..constants import MLFLOW_ARTIFACT_DATA_PATH
from ..model.decomposition import DecompositionWorkflowBase, MDSDecomposition, PCADecomposition, TSNEDecomposition
from ._base import ModelSelectionBase


class DecompositionModelSelection(ModelSelectionBase):
    """Simulate the normal way of invoking scikit-learn decomposition algorithms."""

    def __init__(self, model_name: str) -> None:
        self.model_name = model_name
        self.dcp_workflow = DecompositionWorkflowBase()
        self.transformer_config = {}

    def activate(
        self,
        X: pd.DataFrame,
        y: Optional[pd.DataFrame] = None,
        X_train: Optional[pd.DataFrame] = None,
        X_test: Optional[pd.DataFrame] = None,
        y_train: Optional[pd.DataFrame] = None,
        y_test: Optional[pd.DataFrame] = None,
    ) -> None:
        """Train by Scikit-learn framework."""

        self.dcp_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test)

        if self.model_name == "PCA":
            hyper_parameters = PCADecomposition.manual_hyper_parameters()
            self.dcp_workflow = PCADecomposition(n_components=hyper_parameters["n_components"], svd_solver=hyper_parameters["svd_solver"])
        elif self.model_name == "T-SNE":
            hyper_parameters = TSNEDecomposition.manual_hyper_parameters()
            self.dcp_workflow = TSNEDecomposition(
                n_components=hyper_parameters["n_components"],
                perplexity=hyper_parameters["perplexity"],
                learning_rate=hyper_parameters["learning_rate"],
                n_iter=hyper_parameters["n_iter"],
                early_exaggeration=hyper_parameters["early_exaggeration"],
            )
        elif self.model_name == "MDS":
            hyper_parameters = MDSDecomposition.manual_hyper_parameters()
            self.dcp_workflow = MDSDecomposition(
                n_components=hyper_parameters["n_components"],
                metric=hyper_parameters["metric"],
                n_init=hyper_parameters["n_init"],
                max_iter=hyper_parameters["max_iter"],
            )

        self.dcp_workflow.show_info()

        # Use Scikit-learn style API to process input data
        X_reduced = self.dcp_workflow.fit_transform(X)
        X_reduced = self.dcp_workflow.np2pd(X_reduced, [f"Dimension {i+1}" for i in range(X_reduced.shape[1])])
        self.dcp_workflow.data_upload(X=X)

        # Save the model hyper-parameters
        self.dcp_workflow.save_hyper_parameters(hyper_parameters, self.model_name, os.getenv("GEOPI_OUTPUT_PARAMETERS_PATH"))

        # Common components for every decomposition algorithm
        self.dcp_workflow.common_components()

        # special components of different algorithms
        self.dcp_workflow.special_components(components_num=hyper_parameters["n_components"], reduced_data=X_reduced)

        # Save decomposition result
        self.dcp_workflow.data_save(X_reduced, "X Reduced", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Reduced Data")

        # Save the trained model
        self.dcp_workflow.model_save()



from pydantic import BaseModel


class Dataset(BaseModel):
    id: int
    name: str = None
    # description: str = None
    json_data: str = None
    sequence: int = None
    user_id: int = None

    class Config:
        orm_mode = True


class Diagram(BaseModel):
    id: int
    name: str = None
    # description: str = None
    image: bytes = None
    dataset_id: int = None

    class Config:
        orm_mode = True



import os

import uvicorn
from auth import router as auth_router
from auth import sql_models as auth_models
from data_mining import router as data_mining_router
from data_mining.dash_pipeline import dash_pipeline
from database import engine
from dotenv import load_dotenv
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.wsgi import WSGIMiddleware
from fastapi.openapi.utils import get_openapi

# Load environment variables
load_dotenv()

# Create tables in database
auth_models.Base.metadata.create_all(bind=engine)

# Create FastAPI app
app = FastAPI()

# Add routers
app.include_router(data_mining_router.router)
app.include_router(auth_router.router)

# Get the allowed origins from the environment
allowed_origins = os.getenv("ALLOWED_ORIGINS", "").split(",")

# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=allowed_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/", tags=["root"])
async def read_root() -> dict:
    return {"message": "Welcome to Geochemistry Pi!"}


# Add Dash app
dash_prefix = os.getenv("DASH_REQUESTS_PATHNAME_PREFIX", "/dash/")
dash_app = dash_pipeline(requests_pathname_prefix=dash_prefix)
app.mount(dash_prefix, WSGIMiddleware(dash_app.server))


api_docs = """
Geochemistry π is a Python framework for data-driven geochemistry discovery. It provides an extendable tool and one-stop shop for geochemical data analysis on tabular data.

### Authentication

This API uses OAuth2 with password flow, so you will need to create a user first by using the `/auth/register` endpoint. After creating a user,
 you can login using the `/auth/login` endpoint to get an access token. You can then use the access token to access the other endpoints.
   When clicking on the `Authorize` button, enter the following credentials:

   username: registered email address <br>
   password: registered password
"""


def custom_openapi():
    # Override OpenAPI schema to change field name in Swagger UI
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="Geochemistry Pi",
        version="0.2.1",
        description=api_docs,
        routes=app.routes,
    )

    # openapi_schema["info"]["x-logo"] = {
    #     "url": "https://raw.githubusercontent.com/GeochemistryPi/geochemistrypi/master/docs/logo.png"
    # }

    app.openapi_schema = openapi_schema
    return app.openapi_schema


# Override OpenAPI schema
app.openapi = custom_openapi

if __name__ == "__main__":
    backend_host = os.getenv("BACKEND_HOST", "0.0.0.0")
    backend_port = int(os.getenv("BACKEND_PORT", 8000))
    # Run the app
    uvicorn.run("start_dash_pipeline:app", host=backend_host, port=backend_port, reload=True)



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input, str_input


def gradient_boosting_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Estimators: The number of boosting stages to perform. Gradient boosting is fairly robust to over-fitting so a large number usually results in better performance.")
    print("Please specify the number of boosting stages to perform. A good starting range could be between 50 and 500, such as 100.")
    n_estimators = num_input(SECTION[2], "@N Estimators: ")
    print("Learning Rate: Learning rate shrinks the contribution of each tree by learning_rate. There is a trade-off between learning_rate and n_estimators.")
    print("Please specify the learning rate. A good starting range could be between 0.01 and 0.2, such as 0.1.")
    learning_rate = float_input(0.1, SECTION[2], "@Learning Rate: ")
    print("Max Depth: The maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit.")
    print("Please specify the maximum depth of a tree. A good starting range could be between 5 and 10, such as 4.")
    max_depth = num_input(SECTION[2], "@Max Depth: ")
    print("Min Samples Split: The minimum number of samples required to split an internal node.")
    print("Please specify the minimum number of samples required to split an internal node. A good starting range could be between 2 and 10, such as 2.")
    min_samples_split = num_input(SECTION[2], "@Min Samples Split: ")
    print("Min Samples Leaf: The minimum number of samples required to be at a leaf node.")
    print("Please specify the minimum number. A good starting range could be between 1 and 10, such as 1.")
    min_samples_leaf = num_input(SECTION[2], "@Min Samples Leaf: ")
    print("Max Features: The number of features to consider when looking for the best split.")
    print("Please specify the number of features. A good starting range could be between 1 and the total number of features in the dataset.")
    print("It's recommended to start with a value of sqrt(n_features) or log2(n_features). Make sure to use integer values.")
    max_features = num_input(SECTION[2], "@Max Features: ")
    print("Subsample: The fraction of samples to be used for fitting the individual base learners. If smaller than 1.0 this results in Stochastic Gradient Boosting.")
    print("Please specify the fraction of samples. A good starting range could be between 0.5 and 1.0, such as 1.0.")
    subsample = float_input(1.0, SECTION[2], "@Subsample: ")
    print("Loss: It refers to the loss function to be minimized in each split.")
    print("Please specify the loss function. It is generally recommended to leave it to 'squared_error'.")
    losses = ["squared_error", "absolute_error", "huber", "quantile"]
    loss = str_input(losses, SECTION[2])
    hyper_parameters = {
        "n_estimators": n_estimators,
        "learning_rate": learning_rate,
        "max_depth": max_depth,
        "min_samples_split": min_samples_split,
        "min_samples_leaf": min_samples_leaf,
        "max_features": max_features,
        "subsample": subsample,
        "loss": loss,
    }
    return hyper_parameters



from database import Base
from sqlalchemy import Column, ForeignKey, Integer, LargeBinary, String, Text
from sqlalchemy.orm import relationship


class Dataset(Base):
    __tablename__ = "datasets"

    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, index=True)
    # description = Column(Text)
    json_data = Column(Text)
    sequence = Column(Integer)
    user_id = Column(Integer, ForeignKey("users.id"))

    user = relationship("User", back_populates="datasets")
    diagrams = relationship("Diagram", back_populates="dataset")


class Diagram(Base):
    __tablename__ = "diagrams"

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, index=True)
    # description = Column(Text)
    image = Column(LargeBinary)
    dataset_id = Column(Integer, ForeignKey("datasets.id"))

    dataset = relationship("Dataset", back_populates="diagrams")



# -*- coding: utf-8 -*-
import os
import platform
import subprocess
import threading
from typing import Optional

import typer

from ._version import __version__
from .data_mining.cli_pipeline import cli_pipeline
from .data_mining.constants import WORKING_PATH

app = typer.Typer()

CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))
FRONTEND_PATH = os.path.join(CURRENT_PATH, "frontend")
BACKEND_PATH = os.path.join(CURRENT_PATH, "start_dash_pipeline.py")
PIPELINE_PATH = os.path.join(CURRENT_PATH, "start_cli_pipeline.py")
MLFLOW_STORE_PATH = os.path.join(f"file:{WORKING_PATH}", "geopi_tracking")


def _version_callback(value: bool) -> None:
    """Show Geochemistry Pi version."""
    if value:
        typer.echo(f"Geochemistry π {__version__}")
        raise typer.Exit()


@app.callback()
def main(version: Optional[bool] = typer.Option(None, "--version", "-v", help="Show version.", callback=_version_callback, is_eager=True)) -> None:
    """
    Geochemistry π is an open-sourced highly automated machine learning Python framework for data-driven geochemistry discovery.
    It has the cores components of continous training, machine learning lifecycle management and model serving.
    """
    return


@app.command()
def data_mining(
    data: str = typer.Option("", help="The path of the training data without model inference."),
    training: str = typer.Option("", help="The path of the training data."),
    application: str = typer.Option("", help="The path of the inference data."),
    mlflow: bool = typer.Option(False, help="Start the mlflow server."),
    web: bool = False,
) -> None:
    """Implement the customized automated machine learning pipeline for geochemistry data mining."""

    def start_backend():
        """Start the backend server."""
        start_backend_command = f"python {BACKEND_PATH}"
        subprocess.run(start_backend_command, shell=True)

    def start_frontend():
        """Start the frontend server."""
        start_frontend_command = f"cd {FRONTEND_PATH} && yarn start"
        subprocess.run(start_frontend_command, shell=True)

    def start_mlflow():
        """Start the mlflow server."""
        start_mlflow_command = f"mlflow ui --backend-store-uri {MLFLOW_STORE_PATH} "
        subprocess.run(start_mlflow_command, shell=True)

    if web:
        # Start the backend and frontend in parallel
        backend_thread = threading.Thread(target=start_backend)
        backend_thread.start()
        frontend_thread = threading.Thread(target=start_frontend)
        frontend_thread.start()
        # Wait for the threads to finish
        backend_thread.join()
        frontend_thread.join()
    else:
        # If mlflow is enabled, start the mlflow server, otherwise start the CLI pipeline
        if mlflow:
            # Start mlflow server to track the experiment
            mlflow_thread = threading.Thread(target=start_mlflow)
            mlflow_thread.start()
        else:
            # If the data is provided, start the CLI pipeline with continuous training
            if data:
                cli_pipeline(data)
            # If the training data and inference data are provided, start the CLI pipeline with continuous training and inference
            elif training and application:
                cli_pipeline(training, application)
            # If no data is provided, use built-in data to start the CLI pipeline with continuous training and inference
            else:
                cli_pipeline(training, application)


@app.command()
def web_setup() -> None:
    """Set up the dependency of the web application."""
    my_os = platform.system()
    if my_os == "Windows":
        # Define the command to download and install Yarn on Windows using Chocolatey package manager
        download_yarn = "choco install yarn"
        subprocess.run(download_yarn, shell=True)
        # Define the command to download and install Node.js on Windows using Chocolatey package manager
        download_node = "choco install nodejs"
        subprocess.run(download_node, shell=True)
    elif my_os == "Linux":
        # Define the command to download and install Yarn on Linux using npm
        download_yarn = "apt-get install -y yarn"
        subprocess.run(download_yarn, shell=True)
        # Define the command to download and install Node.js on Linux using npm
        download_node = "apt-get install -y nodejs"
        subprocess.run(download_node, shell=True)
    elif my_os == "Darwin":
        try:
            check_node = "node --version"
            subprocess.run(check_node, shell=True)
            print("Node.js is already installed.")
        except subprocess.CalledProcessError:
            # Define the command to download and install Node.js on macOS using Homebrew
            download_node = "brew install node"
            subprocess.run(download_node, shell=True)
        try:
            # Define the command to check if Yarn is installed
            check_yarn = "yarn --version"
            subprocess.run(check_yarn, shell=True)
            print("Yarn is already installed.")
        except subprocess.CalledProcessError:
            # Define the command to download and install Yarn on macOS using Homebrew
            download_yarn = "brew install yarn"
            subprocess.run(download_yarn, shell=True)

        # Define the command to install the frontend dependencies
        install_frontend_dependency_cmd = f"cd {FRONTEND_PATH} && yarn install"
        subprocess.run(install_frontend_dependency_cmd, shell=True)



# -*- coding: utf-8 -*-
import os
from typing import Optional

import pandas as pd

from ..model.clustering import AffinityPropagationClustering, Agglomerative, ClusteringWorkflowBase, DBSCANClustering, KMeansClustering
from ._base import ModelSelectionBase


class ClusteringModelSelection(ModelSelectionBase):
    """Simulate the normal way of invoking scikit-learn clustering algorithms."""

    def __init__(self, model_name: str) -> None:
        self.model_name = model_name
        self.clt_workflow = ClusteringWorkflowBase()
        self.transformer_config = {}

    def activate(
        self,
        X: pd.DataFrame,
        y: Optional[pd.DataFrame] = None,
        X_train: Optional[pd.DataFrame] = None,
        X_test: Optional[pd.DataFrame] = None,
        y_train: Optional[pd.DataFrame] = None,
        y_test: Optional[pd.DataFrame] = None,
    ) -> None:
        """Train by Scikit-learn framework."""

        self.clt_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test)

        if self.model_name == "KMeans":
            hyper_parameters = KMeansClustering.manual_hyper_parameters()
            self.clt_workflow = KMeansClustering(
                n_clusters=hyper_parameters["n_clusters"],
                init=hyper_parameters["init"],
                max_iter=hyper_parameters["max_iter"],
                tol=hyper_parameters["tol"],
                algorithm=hyper_parameters["algorithm"],
            )
        elif self.model_name == "DBSCAN":
            hyper_parameters = DBSCANClustering.manual_hyper_parameters()
            self.clt_workflow = DBSCANClustering(
                eps=hyper_parameters["eps"],
                min_samples=hyper_parameters["min_samples"],
                metric=hyper_parameters["metric"],
                algorithm=hyper_parameters["algorithm"],
                leaf_size=hyper_parameters["leaf_size"],
                p=hyper_parameters["p"],
            )
        elif self.model_name == "Agglomerative":
            hyper_parameters = Agglomerative.manual_hyper_parameters()
            self.clt_workflow = Agglomerative(
                n_clusters=hyper_parameters["n_clusters"],
                linkage=hyper_parameters["linkage"],
            )
        elif self.model_name == "AffinityPropagation":
            hyper_parameters = AffinityPropagationClustering.manual_hyper_parameters()
            self.clt_workflow = AffinityPropagationClustering(
                damping=hyper_parameters["damping"],
                max_iter=hyper_parameters["max_iter"],
                convergence_iter=hyper_parameters["convergence_iter"],
                affinity=hyper_parameters["affinity"],
            )
        elif self.model_name == "":
            pass

        self.clt_workflow.show_info()

        # Use Scikit-learn style API to process input data
        self.clt_workflow.fit(X)
        self.clt_workflow.get_cluster_centers()
        self.clt_workflow.get_labels()

        # Save the model hyper-parameters
        self.clt_workflow.save_hyper_parameters(hyper_parameters, self.model_name, os.getenv("GEOPI_OUTPUT_PARAMETERS_PATH"))

        # Common components for every clustering algorithm
        self.clt_workflow.common_components()

        # special components of different algorithms
        self.clt_workflow.special_components()

        # Save the trained model
        self.clt_workflow.model_save()



from typing import Union

import mlflow


def retrieve_previous_experiment_id(experiment_name: str) -> Union[str, None]:
    """Retrieve the previous experiment with the same name.

    Parameters
    ----------
    experiment_name : str
        The name of the experiment.

    Returns
    -------
    experiment_id : str
        The ID of the experiment.
    """
    experiment = mlflow.get_experiment_by_name(experiment_name)
    if experiment:
        return experiment.experiment_id
    else:
        return None



# -*- coding: utf-8 -*-
from itertools import cycle
from typing import Dict

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from rich import print
from sklearn.metrics import calinski_harabasz_score, silhouette_samples, silhouette_score


def score(data: pd.DataFrame, labels: pd.DataFrame) -> Dict:
    """Calculate the scores of the clustering model.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
        The true values.

    labels : pd.DataFrame (n_samples, n_components)
        Labels of each point.

    Returns
    -------
    scores : dict
        The scores of the clustering model.
    """
    silhouette = silhouette_score(data, labels)
    calinski_harabaz = calinski_harabasz_score(data, labels)
    print("silhouette_score: ", silhouette)
    print("calinski_harabasz_score:", calinski_harabaz)
    scores = {
        "silhouette_score": silhouette,
        "calinski_harabasz_score": calinski_harabaz,
    }
    return scores


def scatter2d(data: pd.DataFrame, labels: pd.DataFrame, cluster_centers_: np.ndarray, algorithm_name: str) -> None:
    """
    Draw the result-2D diagram for analysis.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
       The features of the data.

    labels : pd.DataFrame (n_samples,)
        Labels of each point.

    cluster_centers_: np.ndarray (n_samples,)
        Coordinates of cluster centers. If the algorithm stops before fully converging (see tol and max_iter), these will not be consistent with labels_.

    algorithm_name : str
        the name of the algorithm
    """
    markers = ["+", "v", ".", "d", "o", "s", "1", "D", "X", "^", "p", "<", "*", "H", "3", "P"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#33a02c",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
    ]

    marker_cycle = cycle(markers)
    color_cycle = cycle(colors)

    fig = plt.figure()
    fig.set_size_inches(18, 10)
    plt.subplot(111)
    # Plot the data
    for i, label in enumerate(set(labels)):
        cluster_data = data[labels == label]
        color = next(color_cycle)
        marker = next(marker_cycle)
        plt.scatter(cluster_data.iloc[:, 0], cluster_data.iloc[:, 1], c=color, marker=marker)

    # Plot the cluster centers
    if not isinstance(cluster_centers_, str):
        # Draw white circles at cluster centers
        plt.scatter(cluster_centers_[:, 0], cluster_centers_[:, 1], c="white", marker="o", alpha=1, s=200, edgecolor="k")

        # Label the cluster centers
        for i, c in enumerate(cluster_centers_):
            plt.scatter(c[0], c[1], marker="$%d$" % i, alpha=1, s=50, edgecolor="k")

    plt.xlabel(f"{data.columns[0]}")
    plt.ylabel(f"{data.columns[1]}")
    plt.title(f"Cluster Data Bi-plot - {algorithm_name}")


def scatter3d(data: pd.DataFrame, labels: pd.DataFrame, algorithm_name: str) -> None:
    """
    Draw the result-3D diagram for analysis.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
       The features of the data.

    labels : pd.DataFrame (n_samples,)
        Labels of each point.

    algorithm_name : str
        the name of the algorithm
    """
    plt.figure()
    namelist = data.columns.values.tolist()
    fig = plt.figure(figsize=(12, 6), facecolor="w")
    plt.subplots_adjust(left=0.05, right=0.95, bottom=0.05, top=0.9)

    # Plot the data without cluster results
    ax = fig.add_subplot(121, projection="3d")
    ax.scatter(data.iloc[:, 0], data.iloc[:, 1], data.iloc[:, 2], alpha=0.3, c="#FF0000", marker=".")
    ax.set_xlabel(namelist[0])
    ax.set_ylabel(namelist[1])
    ax.set_zlabel(namelist[2])
    plt.grid(True)

    ax2 = fig.add_subplot(122, projection="3d")
    markers = ["+", "v", ".", "d", "o", "s", "1", "D", "X", "^", "p", "<", "*", "H", "3", "P"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#33a02c",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
    ]
    marker_cycle = cycle(markers)
    color_cycle = cycle(colors)

    # Plot the data with cluster results
    for i, label in enumerate(set(labels)):
        cluster_data = data[labels == label]
        color = next(color_cycle)
        marker = next(marker_cycle)
        ax2.scatter(cluster_data.iloc[:, 0], cluster_data.iloc[:, 1], cluster_data.iloc[:, 2], c=color, marker=marker, s=6, cmap=plt.cm.Paired, edgecolors="none")

    ax2.set_xlabel(namelist[0])
    ax2.set_ylabel(namelist[1])
    ax2.set_zlabel(namelist[2])
    plt.grid(True)
    ax.set_title(f"Base Data Tri-plot - {algorithm_name}")
    ax2.set_title(f"Cluster Data Tri-plot - {algorithm_name}")


def plot_silhouette_diagram(data: pd.DataFrame, labels: pd.DataFrame, cluster_centers_: np.ndarray, model: object, algorithm_name: str) -> None:
    """
    Draw the silhouette diagram for analysis.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
       The true values.

    labels : pd.DataFrame (n_samples,)
        Labels of each point.

    cluster_centers_: np.ndarray (n_samples,)
        Coordinates of cluster centers. If the algorithm stops before fully converging (see tol and max_iter), these will not be consistent with labels_.

    model : sklearn algorithm model
        The sklearn algorithm model trained with X.

    algorithm_name : str
        the name of the algorithm

    References
    ----------
    Silhouette analysis can be used to study the separation distance between the resulting clusters.
    The silhouette plot displays a measure of how close each point in one cluster is to other points in the
    neighboring clusters and thus provides a way to assess parameters like number of clusters visually.
    This measure has a range of [-1, 1].

    https://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_silhouette_analysis.html
    """
    if hasattr(model, "n_clusters"):
        n_clusters = model.n_clusters
    else:
        n_clusters = len(set(labels)) - (1 if -1 in labels else 0)

    # Create a subplot with 1 row and 2 columns
    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(18, 10)

    # The 1st subplot is the silhouette plot
    # The silhouette coefficient can range from -1, 1 but in this example all
    # lie within [-0.1, 1]
    ax1.set_xlim([-0.1, 1])
    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters, to demarcate them clearly.
    ax1.set_ylim([0, len(data) + (n_clusters + 1) * 10])

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    silhouette_avg = silhouette_score(data, labels)
    print(
        "For n_clusters =",
        n_clusters,
        "The average silhouette_score is :",
        silhouette_avg,
    )

    # Compute the silhouette scores for each sample
    sample_silhouette_values = silhouette_samples(data, labels)

    if n_clusters >= 20:
        Fontsize = 5
        y_long = 7
    else:
        Fontsize = None
        y_long = 10

    y_lower = 10
    for i in range(n_clusters):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = sample_silhouette_values[labels == i]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        color = cm.nipy_spectral(float(i) / n_clusters)
        ax1.fill_betweenx(
            np.arange(y_lower, y_upper),
            0,
            ith_cluster_silhouette_values,
            facecolor=color,
            edgecolor=color,
            alpha=0.7,
        )

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i), fontsize=Fontsize)

        # Compute the new y_lower for next plot
        y_lower = y_upper + y_long  # 10 for the 0 samples

    ax1.set_title("The silhouette plot for the various clusters.")
    ax1.set_xlabel("The silhouette coefficient values")
    ax1.set_ylabel("Cluster label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

    # 2nd Plot showing the actual clusters formed
    colors = cm.nipy_spectral(labels.astype(float) / n_clusters)
    ax2.scatter(data.iloc[:, 0], data.iloc[:, 1], marker=".", s=30, lw=0, alpha=0.7, c=colors, edgecolor="k")

    if not isinstance(cluster_centers_, str):
        # Draw white circles at cluster centers
        ax2.scatter(
            cluster_centers_[:, 0],
            cluster_centers_[:, 1],
            marker="o",
            c="white",
            alpha=1,
            s=200,
            edgecolor="k",
        )

        # Label the cluster centers
        for i, c in enumerate(cluster_centers_):
            ax2.scatter(c[0], c[1], marker="$%d$" % i, alpha=1, s=50, edgecolor="k")

    ax2.set_title("The visualization of the clustered data.")
    ax2.set_xlabel("Feature space for the 1st feature")
    ax2.set_ylabel("Feature space for the 2nd feature")
    plt.suptitle(
        f"Silhouette analysis for clustering on sample data with n_clusters = %d - {algorithm_name}" % n_clusters,
        fontsize=14,
        fontweight="bold",
    )


def plot_silhouette_value_diagram(data, labels, algorithm_name: str):
    """Calculate the scores of the clustering model.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
        The true values.

    labels : pd.DataFrame (n_samples, n_components)
        Labels of each point.

    algorithm_name : str
        The name of the algorithm model.
    """
    silhouette_values = silhouette_samples(data, labels)
    sns.histplot(silhouette_values, bins=30, kde=True)
    plt.title(f"Silhouette value Diagram - {algorithm_name}")
    plt.xlabel("Silhouette Coefficient")
    plt.ylabel("Frequency")



import os

# The number of uploading dataset per user is limited to 5.
MAX_UPLOADS_PER_USER = 5

# the directory in which the package(application) is installed
PACKAGEDIR = os.path.dirname(os.path.realpath(__file__))

# the directory where the built-in data set to be processed stays
BUILT_IN_DATASET_PATH = os.path.join(PACKAGEDIR, "data", "dataset")

# current working directory in which the user activates the application
WORKING_PATH = os.getcwd()

# the root directory where all the output stays
OUTPUT_PATH = os.path.join(WORKING_PATH, "geopi_output")

# the directory where the artifact is saved within the MLflow run's artifact directory
MLFLOW_ARTIFACT_DATA_PATH = "data"
MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH = os.path.join("image", "statistic")
MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH = os.path.join("image", "model_output")
MLFLOW_ARTIFACT_IMAGE_MAP_PATH = os.path.join("image", "map")

# Tell which section the user is currently in on the UML
SECTION = ["User", "Data", "Model", "Plot"]

OPTION = ["Yes", "No"]
DATA_OPTION = ["Own Data", "Testing Data (Built-in)"]
TEST_DATA_OPTION = ["Data For Regression", "Data For Classification", "Data For Clustering", "Data For Dimensional Reduction", "Data For Abnormal Detection"]
MODE_OPTION = ["Regression", "Classification", "Clustering", "Dimensional Reduction", "Abnormal Detection"]
MODE_OPTION_WITH_MISSING_VALUES = ["Regression", "Classification", "Clustering"]

# The model provided to use
REGRESSION_MODELS = [
    "Linear Regression",
    "Polynomial Regression",
    "K-Nearest Neighbors",
    "Support Vector Machine",
    "Decision Tree",
    "Random Forest",
    "Extra-Trees",
    "Gradient Boosting",
    "XGBoost",
    "Multi-layer Perceptron",
    "Lasso Regression",
    "Elastic Net",
    "SGD Regression",
    "BayesianRidge Regression",
    "Ridge Regression",
    # "Bagging Regression",
    # "Decision Tree",
    # Histogram-based Gradient Boosting,
]
CLASSIFICATION_MODELS = [
    "Logistic Regression",
    "Support Vector Machine",
    "Decision Tree",
    "Random Forest",
    "Extra-Trees",
    "XGBoost",
    "Multi-layer Perceptron",
    "Gradient Boosting",
    "K-Nearest Neighbors",
    "Stochastic Gradient Descent",
    # "Bagging Classification",
    # "Decision Tree",
    # Histogram-based Gradient Boosting,
]
CLUSTERING_MODELS = ["KMeans", "DBSCAN", "Agglomerative", "AffinityPropagation"]
DECOMPOSITION_MODELS = ["PCA", "T-SNE", "MDS"]
ABNORMALDETECTION_MODELS = ["Isolation Forest"]

# The model can deal with missing values
# Reference: https://scikit-learn.org/stable/modules/impute.html#estimators-that-handle-nan-values
REGRESSION_MODELS_WITH_MISSING_VALUES = [
    "XGBoost",
    # "Bagging Regression",
    # "Decision Tree",
    # Histogram-based Gradient Boosting,
]
CLASSIFICATION_MODELS_WITH_MISSING_VALUES = [
    "XGBoost",
    # "Bagging Classification",
    # "Decision Tree",
    # Histogram-based Gradient Boosting,
]
CLUSTERING_MODELS_WITH_MISSING_VALUES = [
    # "HDBSCAN"
]


# Special AutoML models
NON_AUTOML_MODELS = ["Linear Regression", "Polynomial Regression"]
RAY_FLAML = ["Multi-layer Perceptron"]

MISSING_VALUE_STRATEGY = ["Drop Rows with Missing Values ", "Impute Missing Values"]

IMPUTING_STRATEGY = ["Mean Value", "Median Value", "Most Frequent Value", "Constant(Specified Value)"]

FEATURE_SCALING_STRATEGY = ["Min-max Scaling", "Standardization"]

SAMPLE_BALANCE_STRATEGY = ["Over Sampling", "Under Sampling", "Oversampling and Undersampling"]

CUSTOMIZE_LABEL_STRATEGY = ["Automatic Coding", "Custom Numeric Labels", "Custom Non-numeric Labels"]

FEATURE_SELECTION_STRATEGY = ["Generic Univariate Select", "Select K Best"]

CALCULATION_METHOD_OPTION = ["Micro", "Macro", "Weighted"]

DROP_MISSING_VALUE_STRATEGY = ["Drop All Rows with Missing Values", "Drop Rows with Missing Values by Specific Columns"]



import auth.sql_models as auth_models
from fastapi import HTTPException
from sqlalchemy.orm import Session

from .constants import MAX_UPLOADS_PER_USER
from .sql_models import Dataset


def read_all_datasets(db: Session, user_id: int):
    return db.query(Dataset).filter_by(user_id=user_id).order_by(Dataset.sequence).all()


def read_basic_datasets_info(db: Session, user_id: int):
    # Select id, name, sequence from datasets where user_id = user_id order by sequence
    basic_datasets_info_list = db.query(Dataset.id, Dataset.name, Dataset.sequence).filter_by(user_id=user_id).order_by(Dataset.sequence).all()
    # Convert the list of tuples to a list of dictionaries
    basic_datasets_info = [dict(zip(["id", "name", "sequence"], row)) for row in basic_datasets_info_list]
    return basic_datasets_info


def read_dataset(db: Session, user_id: int, dataset_id: int):
    return db.query(Dataset).filter_by(user_id=user_id, id=dataset_id).first()


def upload_dataset(db: Session, user_id: int, json_dataset: str, dataset_name: str):
    user = db.query(auth_models.User).get(user_id)
    if user.upload_count >= MAX_UPLOADS_PER_USER:
        raise HTTPException(status_code=429, detail="User has reached maximum number of uploads")

    # Calculate the new sequence number for the dataset
    new_sequence = user.upload_count + 1

    # Update the sequence numbers of existing datasets
    existing_datasets = db.query(Dataset).filter_by(user_id=user_id).order_by(Dataset.sequence)
    for dataset in existing_datasets:
        if dataset.sequence >= new_sequence:
            dataset.sequence += 1

    # Create a new Dataset instance and associate it with the user
    new_dataset = Dataset(json_data=json_dataset, sequence=new_sequence, name=dataset_name, user_id=user_id)
    db.add(new_dataset)
    # Commit the new dataset to the database
    db.commit()
    # Refresh the dataset to get the id
    db.refresh(new_dataset)

    # Update the user's upload count
    user.upload_count += 1
    db.commit()

    # Update associated diagrams
    # update_diagrams(db, new_dataset.id, json_dataset)

    return new_dataset


def remove_dataset(db: Session, user_id: int, dataset_id: int):
    dataset = db.query(Dataset).get(dataset_id)
    if not dataset:
        raise Exception("Dataset not found")

    # Delete associated diagrams
    # delete_diagrams(db, dataset_id)

    # Update the sequence numbers of existing datasets
    existing_datasets = db.query(Dataset).filter_by(user_id=user_id).order_by(Dataset.sequence)
    for other_dataset in existing_datasets:
        if other_dataset.sequence > dataset.sequence:
            other_dataset.sequence -= 1

    # Delete the dataset
    db.delete(dataset)
    db.commit()

    # Update the user's upload count
    user = db.query(auth_models.User).get(user_id)
    user.upload_count -= 1
    db.commit()

    return dataset



import os

from dotenv import load_dotenv

load_dotenv()

# to get a string like this run:
# openssl rand -hex 32
SECRET_KEY = "5277ee28edc76a2810312a90ab8ab53a6d9fa16be8a4db446ad93e66b2444d02"
ALGORITHM = os.getenv("ALGORITHM", "HS256")
ACCESS_TOKEN_EXPIRE_MINUTES = int(os.getenv("ACCESS_TOKEN_EXPIRE_MINUTES", 30))



# -*- coding: utf-8 -*-
from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input, str_input, tuple_input


def multi_layer_perceptron_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("Hidden Layer Sizes: The ith element represents the number of neurons in the ith hidden layer.")
    print("Please specify the size of hidden layer and the number of neurons in the each hidden layer.")
    hidden_layer = tuple_input((50, 25, 5), SECTION[2], "@Hidden Layer Sizes: ")
    print("Activation: Activation function for the hidden layer.")
    print("Please specify the activation function for the hidden layer. It is generally recommended to leave it as 'ReLU'.")
    activations = ["identity", "logistic", "tanh", "relu"]
    activation = str_input(activations, SECTION[2])
    print("Solver: The solver for weight optimization.")
    print("Please specify the solver for weight optimization. It is generally recommended to leave it as 'Adam'.")
    solvers = ["lbfgs", "sgd", "adam"]
    solver = str_input(solvers, SECTION[2])
    print("Alpha: L2 penalty (regularization term) parameter.")
    print("Please specify the L2 penalty (regularization term) parameter. A good starting range could be between 0.0001 and 10, such as 0.0001.")
    alpha = float_input(0.0001, SECTION[2], "@Alpha: ")
    print("Learning Rate: It controls the step-size in updating the weights.")
    print("Please specify the learning rate. It is generally recommended to leave it as 'Adaptive'.")
    learning_rates = ["constant", "invscaling", "adaptive"]
    learning_rate = str_input(learning_rates, SECTION[2])
    print("Max Iterations: Maximum number of iterations. The solver iterates until convergence (determined by 'tol') or this number of iterations.")
    print("Please specify the maximum number of iterations. A good starting range could be between 100 and 1000, such as 200.")
    max_iter = num_input(SECTION[2], "@Max Iterations: ")
    hyper_parameters = {
        "hidden_layer_sizes": hidden_layer,
        "activation": activation,
        "solver": solver,
        "alpha": alpha,
        "learning_rate": learning_rate,
        "max_iter": max_iter,
    }
    return hyper_parameters



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, float_input, num_input, str_input


def sgd_classificaiton_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters for SGD Classification.

    Returns
    -------
    hyper_parameters : dict
        The hyperparameters.
    """
    print("Loss Function: The loss function to be used. Choose from 'log_loss', 'modified_huber'.")
    losses = ["log_loss", "modified_huber"]
    loss = str_input(losses, SECTION[2])

    print("Penalty: The penalty (regularization term) to be used. Choose from 'l2', 'l1', 'elasticnet', or None. The default is 'l2'.")
    penalties = ["l2", "l1", "elasticnet", "None"]
    penalty = str_input(penalties, SECTION[2])
    if penalty == "None":
        penalty = None

    l1_ratio = None
    if penalty == "elasticnet":
        print("L1 Ratio: The Elastic Net mixing parameter, with 0 <= l1_ratio <= 1. l1_ratio=0 corresponds to L2 penalty, l1_ratio=1 to L1. The default is 0.15.")
        l1_ratio = float_input(0.15, SECTION[2], "@L1 Ratio: ")

    print("Alpha: The regularization strength. A good starting value could be between 0.0001 and 1, such as 0.01. The default is 0.0001.")
    alpha = float_input(0.0001, SECTION[2], "@Alpha: ")

    print("Fit Intercept: Whether to fit an intercept term. It is generally recommended to set it as True.")
    fit_intercept = bool_input(SECTION[2])

    print("Maximum Number of Iterations: The maximum number of iterations for training. The default is 1000.")
    max_iter = num_input(SECTION[2], "@Max Iterations: ")

    print("Tolerance: The tolerance for stopping criteria. A good starting value could be between 0.000001 and 0.001, such as 0.0001. The default is 0.001.")
    tol = float_input(0.001, SECTION[2], "@Tolerance: ")

    print("Shuffle: Whether or not to shuffle the training data after each epoch. The default is 'True'.")
    shuffle = bool_input(SECTION[2])

    print("Learning Rate: The learning rate schedule. Choose from 'constant', 'optimal', 'invscaling', or 'adaptive'. The default is 'optimal'.")
    learning_rates = ["constant", "optimal", "invscaling", "adaptive"]
    learning_rate = str_input(learning_rates, SECTION[2])

    print("Initial Learning Rate: The initial learning rate for the 'constant', 'invscaling' or 'adaptive' schedules. The default value is 0.0.")
    eta0 = float_input(0.0, SECTION[2], "@Initial Learning Rate: ")

    print("Power T: The exponent for inverse scaling learning rate. The default is 0.5.")
    power_t = float_input(0.5, SECTION[2], "@Power T: ")

    print("Early Stopping: Whether to use early stopping to terminate training when validation score is not improving. The default is 'False'.")
    early_stopping = bool_input(SECTION[2])

    print("Validation Fraction: The proportion of training data to set aside as validation set for early stopping.")
    print("It must be in range (0, 1). A good starting value could be between 0.000001 and 1, such as 0.1. The default is 0.1.")
    is_valid = False
    while not is_valid:
        validation_fraction = float_input(0.1, SECTION[2], "@Validation Fraction: ")
        if 0 < validation_fraction < 1:
            is_valid = True
        else:
            print("The validation fraction must be in range (0, 1).")
    print("Number of Iterations With No Improvement: Number of iterations with no improvement to wait before stopping fitting.")
    print("A good starting value could be between 1 and maximum number of iterations, such as 5. The default is 5.")
    n_iter_no_change = num_input(SECTION[2], "@Iterations With No Improvement: ")

    hyper_parameters = {
        "loss": loss,
        "penalty": penalty,
        "alpha": alpha,
        "fit_intercept": fit_intercept,
        "max_iter": max_iter,
        "tol": tol,
        "shuffle": shuffle,
        "learning_rate": learning_rate,
        "eta0": eta0,
        "power_t": power_t,
        "early_stopping": early_stopping,
        "validation_fraction": validation_fraction,
        "n_iter_no_change": n_iter_no_change,
    }
    if not l1_ratio:
        # Use the default value provided by sklearn.ensemble.SGDClassifier.
        hyper_parameters["l1_ratio"] = 0.15
    else:
        hyper_parameters["l1_ratio"] = l1_ratio

    return hyper_parameters



# -*- coding: utf-8 -*-
import json
import os
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

import mlflow
import numpy as np
import pandas as pd
from multipledispatch import dispatch
from rich import print

from ..constants import SECTION
from ..data.data_readiness import limit_num_input, num2option, num_input, show_data_columns
from ..utils.base import save_data, save_fig, save_model, save_text
from .func._common_supervised import plot_decision_tree, plot_feature_importance, plot_permutation_importance, show_formula
from .func.algo_regression._linear_regression import plot_2d_line_diagram, plot_2d_scatter_diagram, plot_3d_scatter_diagram, plot_3d_surface_diagram


class WorkflowBase(metaclass=ABCMeta):
    """Base class for all workflow classes in geochemistry π."""

    # Default for child class. They need to be overwritten in child classes.
    name = None
    common_function = []
    special_function = []
    X, y = None, None
    X_train, X_test, y_train, y_test = None, None, None, None
    y_test_predict = None

    @classmethod
    def show_info(cls) -> None:
        """Display how many functions the algorithm will provide."""
        print("*-*" * 2, cls.name, "is running ...", "*-*" * 2)
        print("Expected Functionality:")
        function = cls.common_function + cls.special_function
        for i in range(len(function)):
            print("+ ", function[i])

    def __init__(self) -> None:
        # Default for child class. They need to be overwritten in child classes.
        self.model = None
        self.naming = None
        self.automl = None
        self.ray_best_model = None
        self.random_state = 42

    @property
    def image_config(self):
        return {
            # Picture layout
            "width": 16,  # number of subgraph rows
            "height": 9,  # number of subgraph columns
            "dpi": 360,  # resolution
            # Main content
            "cmap": "coolwarm_r",  # color setting
            "cmap2": "Wistia",
            "marker_angle": "^",  # point type
            "marker_circle": "o",  # point type
            "edgecolor": "w",  # point edge color
            "markersize1": 18,  # point size
            "markersize2": 6,
            "alpha1": 0.4,  # point transparency
            "alpha2": 0.95,
            "linestyle": "-",
            # bar
            "bar_color": "blue",
            "bar_align": "center",
            "bar_x": range(len(self.X.columns)),  # the sequence of horizontal coordinates of the bar
            "bar_height": None,  # the height(s) of the bars
            "bar_label": self.X.columns,  # The label on the X-axis
            "bar_width": 0.3,  # the width(s) of the bars
            "bottom": 0,  # the y coordinate(s) of the bars bases
            # Convert the font of the axes
            "labelsize": 5,  # the font size of the axis label
            "xrotation": 0,  # x axis label rotation Angle
            "xha": "center",  # x axis 'ha'
            "rot": 90,  # y axis label rotation Angle
            "yha": "center",  # y axis 'ha'
            "axislabelfont": "Times New Roman",  # axis label font
            # Picture title adjustment
            "title_label": self.naming,  # picture name
            "title_size": 15,  # title font size
            "title_color": "k",
            "title_location": "center",
            "title_font": "Times New Roman",
            "title_pad": 2,
            # Tree parameter
            "max_depth": None,  # The maximum depth of the representation
            "feature_names": None,  # Names of each of the features
            "class_names": ["class" + str(i) for i in range(1, 1000)],  # Names of each of the target classes in ascending numerical order
            "label": "all",  # Whether to show informative labels for impurity, etc
            "filled": True,  # color filling
            "impurity": True,  # When set to True, show the impurity at each node
            "node_ids": None,  # When set to True, show the ID number on each node
            "proportion": False,  # When set to True, change the display of ‘values’ and/or ‘samples’ to be proportions and percentages respectively
            "rounded": True,  # When set to True, draw node boxes with rounded corners and use Helvetica fonts instead of Times-Roman
            "precision": 3,  # Number of digits of precision for floating point in the values of impurity, threshold and value attributes of each node
            "ax": None,  # axes to plot to.
            "fontsize": None,  # size of text font
        }

    @abstractmethod
    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> None:
        """Placeholder for fit. child classes should implement this method!

        Parameters
        ----------
        X : pd.DataFrame (n_samples, n_features)
            Training data, where `n_samples` is the number of samples and n_features` is the number of features.

        y : pd.DataFrame, default=None (n_samples,) or (n_samples, n_targets)
            Target values。
        """
        return None

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """The interface for the child classes."""
        return pd.DataFrame()

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """The interface for the child classes."""
        return pd.DataFrame()

    @abstractmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Placeholder for manual_hyper_parameters. child classes should implement this method!

        Parameters
        ----------
        kwargs : dict
            The hyper parameters of the model.
        """
        return dict()

    @staticmethod
    def score(y_true: Union[pd.DataFrame, np.ndarray], y_predict: Union[pd.DataFrame, np.ndarray]) -> Union[int, float]:
        """The interface for the child classes."""
        return float()

    @staticmethod
    def np2pd(array: np.ndarray, columns_name: Union[List[str], pd.Index]) -> pd.DataFrame:
        """The type of the data set is transformed from numpy.ndarray to pandas.DataFrame.

        Parameters
        ----------
        array : np.ndarray (n_samples, n_features)
            the data set.

        columns_name : list[str] or pd.Index
            the name of the columns of the data set.

        Returns
        -------
        pd.DataFrame (n_samples, n_features)
            the data set.
        """
        return pd.DataFrame(array, columns=columns_name)

    @staticmethod
    def choose_dimension_data(data: pd.DataFrame, dimensions: int) -> Tuple[List[int], pd.DataFrame]:
        """Choose a subgroup data from the whole data set to draw 2d or 3d graph.

        Parameters
        ----------
        data : pd.DataFrame (n_samples, n_features)
            the whole data.
        dimensions : int
            how much dimensions data to keep.

        Returns
        -------
        selected_axis_index : list[int]
            the index of the that dimension, which is shown as the index of the column of the data set.

        selected_axis_data : pd.DataFrame (n_samples, n_features)
            the selected data from the whole data set.
        """
        print(f"-----* {dimensions} Dimensions Data Selection *-----")
        print(f"The software is going to draw related {dimensions}d graphs.")
        print(f"Currently, the data dimension is beyond {dimensions} dimensions.")
        print(f"Please choose {dimensions} dimensions of the data below.")
        data = pd.DataFrame(data)
        selected_axis_index = []
        selected_axis_name = []
        for i in range(1, dimensions + 1):
            num2option(data.columns)
            print(f"Choose dimension - {i} data:")
            index_axis = limit_num_input(data.columns, SECTION[3], num_input)
            selected_axis_index.append(index_axis - 1)
            selected_axis_name.append(data.columns[index_axis - 1])
        selected_axis_data = data.loc[:, selected_axis_name]
        print("The Selected Data Dimension:")
        show_data_columns(selected_axis_name)
        return selected_axis_index, selected_axis_data

    @staticmethod
    def data_upload(
        X: Optional[pd.DataFrame] = None,
        y: Optional[pd.DataFrame] = None,
        X_train: Optional[pd.DataFrame] = None,
        X_test: Optional[pd.DataFrame] = None,
        y_train: Optional[pd.DataFrame] = None,
        y_test: Optional[pd.DataFrame] = None,
        y_train_predict: Optional[pd.DataFrame] = None,
        y_test_predict: Optional[pd.DataFrame] = None,
    ) -> None:
        """This method loads the required data into the base class's attributes."""
        if X is not None:
            WorkflowBase.X = X
        if y is not None:
            WorkflowBase.y = y
        if X_train is not None:
            WorkflowBase.X_train = X_train
        if X_test is not None:
            WorkflowBase.X_test = X_test
        if y_train is not None:
            WorkflowBase.y_train = y_train
        if y_test is not None:
            WorkflowBase.y_test = y_test
        if y_test_predict is not None:
            WorkflowBase.y_test_predict = y_test_predict
        if y_train_predict is not None:
            WorkflowBase.y_train_predict = y_train_predict

    @staticmethod
    def data_save(df: pd.DataFrame, df_name: str, local_path: str, mlflow_path: str, slogan: str) -> None:
        """This method saves the data into the local path and the mlflow path.

        Parameters
        ----------
        df : pd.DataFrame
            The data to be saved.

        df_name : str
            The name of the data.

        local_path : str
            The local path to save the data.

        mlflow_path : str
            The mlflow path to save the data.

        slogan : str
            The title of the output section.
        """
        print(f"-----* {slogan} *-----")
        print(df)
        save_data(df, df_name, local_path, mlflow_path)

    @staticmethod
    def save_hyper_parameters(hyper_parameters_dict: Dict, model_name: str, local_path: str) -> None:
        """This method saves the hyper parameters into the local path.

        Parameters
        ----------
        hyper_parameters_dict : dict
            The hyper parameters of the model.

        model_name : str
            The name of the model.

        local_path : str
            The local path to save the hyper parameters.
        """
        hyper_parameters_str = json.dumps(hyper_parameters_dict, indent=4)
        save_text(hyper_parameters_str, f"Hyper Parameters - {model_name}", local_path)
        mlflow.log_params(hyper_parameters_dict)

    @dispatch()
    def model_save(self) -> None:
        """Persist the model for future use after training the model with Scikit-learn framework."""
        print("-----* Model Persistence *-----")
        GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH")
        save_model(self.model, self.naming, self.X_train.iloc[[0]], GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH)

    @dispatch(bool)
    def model_save(self, is_automl: bool) -> None:
        """Persist the model for future use after training the model with FLAML framework."""
        print("-----* Model Persistence *-----")
        GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH")
        save_model(self.auto_model, self.naming, self.X_train.iloc[[0]], GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH)

    @staticmethod
    def _plot_permutation_importance(
        X_test: pd.DataFrame,
        y_test: pd.DataFrame,
        trained_model: object,
        image_config: dict,
        algorithm_name: str,
        local_path: str,
        mlflow_path: str,
    ) -> None:
        """Permutation importance plot."""
        print("-----* Permutation Importance Diagram *-----")
        importances_mean, importances_std, importances = plot_permutation_importance(X_test, y_test, trained_model, image_config)
        save_fig(f"Permutation Importance - {algorithm_name}", local_path, mlflow_path)
        save_data(X_test, "Permutation Importance - X Test", local_path, mlflow_path)
        save_data(y_test, "Permutation Importance - Y Test", local_path, mlflow_path)
        data_dict = {"importances_mean": importances_mean.tolist(), "importances_std": importances_std.tolist(), "importances": importances.tolist()}
        data_str = json.dumps(data_dict, indent=4)
        save_text(data_str, f"Permutation Importance - {algorithm_name}", local_path, mlflow_path)


class TreeWorkflowMixin:
    """Mixin class for tree models."""

    @staticmethod
    def _plot_feature_importance(X_train: pd.DataFrame, trained_model: object, image_config: dict, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Draw the feature importance bar diagram."""
        print("-----* Feature Importance Diagram *-----")
        columns_name = X_train.columns
        feature_importances = trained_model.feature_importances_
        data = plot_feature_importance(columns_name, feature_importances, image_config)
        save_fig(f"Feature Importance - {algorithm_name}", local_path, mlflow_path)
        save_data(data, f"Feature Importance - {algorithm_name}", local_path, mlflow_path, True)

    @staticmethod
    def _plot_tree(trained_model: object, image_config: dict, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Drawing decision tree diagrams."""
        print("-----* Single Tree Diagram *-----")
        plot_decision_tree(trained_model, image_config)
        save_fig(f"Tree Diagram - {algorithm_name}", local_path, mlflow_path)


class LinearWorkflowMixin:
    """Mixin class for linear models."""

    @staticmethod
    def _show_formula(
        coef: np.ndarray, intercept: np.ndarray, features_name: np.ndarray, algorithm_name: str, regression_classification: str, y_train: pd.DataFrame, local_path: str, mlflow_path: str
    ) -> None:
        """Show the formula."""
        print(f"-----* {algorithm_name} Formula *-----")
        formula = show_formula(coef, intercept, features_name, regression_classification, y_train)
        formula_str = json.dumps(formula, indent=4)
        save_text(formula_str, f"{algorithm_name} Formula", local_path, mlflow_path)

    @staticmethod
    def _plot_2d_scatter_diagram(feature_data: pd.DataFrame, target_data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the 2D graph of the linear regression model."""
        print("-----* 2D Scatter Diagram *-----")
        plot_2d_scatter_diagram(feature_data, target_data)
        save_fig(f"2D Scatter Diagram - {algorithm_name}", local_path, mlflow_path)
        data = pd.concat([feature_data, target_data], axis=1)
        save_data(data, f"2D Scatter Diagram - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _plot_2d_line_diagram(feature_data: pd.DataFrame, target_data: pd.DataFrame, y_test_predict: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the 2D graph of the linear regression model."""
        print("-----* 2D Line Diagram *-----")
        plot_2d_line_diagram(feature_data, target_data, y_test_predict)
        save_fig(f"2D Line Diagram - {algorithm_name}", local_path, mlflow_path)
        data = pd.concat([feature_data, target_data, y_test_predict], axis=1)
        save_data(data, f"2D Line Diagram - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _plot_3d_scatter_diagram(feature_data: pd.DataFrame, target_data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the 3D graph of the linear regression model."""
        print("-----*  3D Scatter Diagram *-----")
        plot_3d_scatter_diagram(feature_data, target_data)
        save_fig(f"3D Scatter Diagram - {algorithm_name}", local_path, mlflow_path)
        data = pd.concat([feature_data, target_data], axis=1)
        save_data(data, f"3D Scatter Diagram - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _plot_3d_surface_diagram(feature_data: pd.DataFrame, target_data: pd.DataFrame, y_test_predict: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the 3D graph of the linear regression model."""
        print("-----* 3D Surface Diagram *-----")
        plot_3d_surface_diagram(feature_data, target_data, y_test_predict)
        save_fig(f"3D Surface Diagram - {algorithm_name}", local_path, mlflow_path)
        data = pd.concat([feature_data, target_data, y_test_predict], axis=1)
        save_data(data, f"3D Surface Diagram - {algorithm_name}", local_path, mlflow_path)



# -*- coding: utf-8 -*-

import os
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from rich import print
from sklearn.decomposition import PCA
from sklearn.manifold import MDS, TSNE

from ..constants import MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH
from ..utils.base import clear_output, save_data, save_fig
from ._base import WorkflowBase
from .func.algo_decomposition._common import plot_2d_scatter_diagram, plot_contour, plot_heatmap
from .func.algo_decomposition._mds import mds_manual_hyper_parameters
from .func.algo_decomposition._pca import biplot, pca_manual_hyper_parameters, triplot
from .func.algo_decomposition._tsne import tsne_manual_hyper_parameters


class DecompositionWorkflowBase(WorkflowBase):
    """The base workflow class of decomposition algorithms."""

    common_function = ["Model Persistence"]  # 'Decomposition Result',

    def __init__(self) -> None:
        super().__init__()

        # the extra attributes that decomposition algorithm needs
        self.X_reduced = None
        self.mode = "Decomposition"

    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> None:
        """Fit the model."""
        self.model.fit(X)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Apply dimensionality reduction to X."""
        return self.model.transform(X)

    def fit_transform(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Fit the model with X and apply the dimensionality reduction on X."""
        # TODO(Can He(Sany) sanhe1097618435@163.com): check if we need to put this function in the base class
        return self.model.fit_transform(X)

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        return dict()

    def _reduced_data2pd(self, reduced_data: np.ndarray, components_num: int) -> None:
        """Transform reduced data into the format of pd.DataFrame.

        Parameters
        ----------
        reduced_data : np.ndarray
            The data X after dimension reduction.

        components_num : int
            The numbers of the principal components.
        """
        pa_name = []
        for i in range(components_num):
            pa_name.append(f"Principal Axis {i+1}")
        self.X_reduced = pd.DataFrame(reduced_data)
        self.X_reduced.columns = pa_name

    @staticmethod
    def _plot_2d_scatter_diagram(data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot the two-dimensional diagram of the decomposition result."""
        print("-----* Decomposition Two-Dimensional Diagram *-----")
        plot_2d_scatter_diagram(data, algorithm_name)
        save_fig(f"Decomposition Two-Dimensional Diagram - {algorithm_name}", local_path, mlflow_path)
        save_data(data, f"Decomposition Two-Dimensional Data - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _plot_heatmap(data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot a heatmap for the decomposition result."""
        print("-----* Decomposition Heatmap *-----")
        plot_heatmap(data, algorithm_name)
        save_fig(f"Decomposition Heatmap - {algorithm_name}", local_path, mlflow_path)
        save_data(data, f"Decomposition Heatmap Data - {algorithm_name}", local_path, mlflow_path)

    @staticmethod
    def _plot_contour(data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Plot a contour plot for dimensionality reduction results."""
        print("-----* Dimensionality Reduction Contour Plot *-----")
        plot_contour(data, algorithm_name)
        save_fig(f"Dimensionality Reduction Contour Plot - {algorithm_name}", local_path, mlflow_path)
        save_data(data, f"Dimensionality Reduction Contour Plot Data - {algorithm_name}", local_path, mlflow_path)

    def common_components(self) -> None:
        """Invoke all common application functions for decomposition algorithms by Scikit-learn framework."""

        GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH")
        self._plot_2d_scatter_diagram(
            data=self.X,
            algorithm_name=self.naming,
            local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
            mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
        )
        self._plot_heatmap(
            data=self.X,
            algorithm_name=self.naming,
            local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
            mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
        )
        self._plot_contour(
            data=self.X,
            algorithm_name=self.naming,
            local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
            mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
        )


class PCADecomposition(DecompositionWorkflowBase):
    """The automation workflow of using PCA algorithm to make insightful products."""

    name = "PCA"
    special_function = ["Principal Components", "Explained Variance Ratio", "Compositional Bi-plot", "Compositional Tri-plot"]

    def __init__(
        self,
        n_components: Optional[int] = None,
        copy: bool = True,
        whiten: bool = False,
        svd_solver: str = "auto",
        tol: float = 0.0,
        iterated_power: Union[int, str] = "auto",
        n_oversamples: int = 10,
        power_iteration_normalizer: str = "auto",
        random_state: Optional[int] = None,
    ) -> None:
        """
        Parameters
        ----------
        n_components : int, float or 'mle', default=None
            Number of components to keep.
            if n_components is not set all components are kept::

                n_components == min(n_samples, n_features)

            If ``n_components == 'mle'`` and ``svd_solver == 'full'``, Minka's
            MLE is used to guess the dimension. Use of ``n_components == 'mle'``
            will interpret ``svd_solver == 'auto'`` as ``svd_solver == 'full'``.

            If ``0 < n_components < 1`` and ``svd_solver == 'full'``, select the
            number of components such that the amount of variance that needs to be
            explained is greater than the percentage specified by n_components.

            If ``svd_solver == 'arpack'``, the number of components must be
            strictly less than the minimum of n_features and n_samples.

            Hence, the None case results in::

                n_components == min(n_samples, n_features) - 1

        copy : bool, default=True
            If False, data passed to fit are overwritten and running
            fit(X).transform(X) will not yield the expected results,
            use fit_transform(X) instead.

        whiten : bool, default=False
            When True (False by default) the `components_` vectors are multiplied
            by the square root of n_samples and then divided by the singular values
            to ensure uncorrelated outputs with unit component-wise variances.
            Whitening will remove some information from the transformed signal
            (the relative variance scales of the components) but can sometime
            improve the predictive accuracy of the downstream estimators by
            making their data respect some hard-wired assumptions.

        svd_solver : {'auto', 'full', 'arpack', 'randomized'}, default='auto'
            If auto :
                The solver is selected by a default policy based on `X.shape` and
                `n_components`: if the input data is larger than 500x500 and the
                number of components to extract is lower than 80% of the smallest
                dimension of the data, then the more efficient 'randomized'
                method is enabled. Otherwise the exact full SVD is computed and
                optionally truncated afterwards.
            If full :
                run exact full SVD calling the standard LAPACK solver via
                `scipy.linalg.svd` and select the components by postprocessing
            If arpack :
                run SVD truncated to n_components calling ARPACK solver via
                `scipy.sparse.linalg.svds`. It requires strictly
                0 < n_components < min(X.shape)
            If randomized :
                run randomized SVD by the method of Halko et al.

            .. versionadded:: 0.18.0

        tol : float, default=0.0
            Tolerance for singular values computed by svd_solver == 'arpack'.
            Must be of range [0.0, infinity).

            .. versionadded:: 0.18.0

        iterated_power : int or 'auto', default='auto'
            Number of iterations for the power method computed by
            svd_solver == 'randomized'.
            Must be of range [0, infinity).

            .. versionadded:: 0.18.0

        n_oversamples : int, default=10
            This parameter is only relevant when `svd_solver="randomized"`.
            It corresponds to the additional number of random vectors to sample the
            range of `X` so as to ensure proper conditioning. See
            :func:`~sklearn.utils.extmath.randomized_svd` for more details.

            .. versionadded:: 1.1

        power_iteration_normalizer : {'auto', 'QR', 'LU', 'none'}, default='auto'
            Power iteration normalizer for randomized SVD solver.
            Not used by ARPACK. See :func:`~sklearn.utils.extmath.randomized_svd`
            for more details.

            .. versionadded:: 1.1

        random_state : int, RandomState instance or None, default=None
            Used when the 'arpack' or 'randomized' solvers are used. Pass an int
            for reproducible results across multiple function calls.
            See :term:`Glossary <random_state>`.

            .. versionadded:: 0.18.0

        References
        ----------
        Scikit-learn API: sklearn.decomposition.PCA
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """
        super().__init__()
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        # self.n_oversamples = n_oversamples
        # self.power_iteration_normalizer = power_iteration_normalizer

        if random_state:
            self.random_state = random_state

        # If 'random_state' is None, 'self.random_state' comes from the parent class 'WorkflowBase'
        self.model = PCA(
            n_components=self.n_components,
            copy=self.copy,
            whiten=self.whiten,
            svd_solver=self.svd_solver,
            tol=self.tol,
            iterated_power=self.iterated_power,
            # n_oversamples=self.n_oversamples,
            # power_iteration_normalizer=self.power_iteration_normalizer,
            random_state=self.random_state,
        )

        self.naming = PCADecomposition.name

        # special attributes
        self.pc_data = None

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = pca_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def _get_principal_components(self) -> None:
        """Get principal components."""
        print("-----* Principal Components *-----")
        print("Every column represents one principal component respectively.")
        print("Every row represents how much that row feature contributes to each principal component respectively.")
        print("The tabular data looks like in format: 'rows x columns = 'features x principal components'.")
        pc_name = []
        for i in range(self.n_components):
            pc_name.append(f"PC{i+1}")
        self.pc_data = pd.DataFrame(self.model.components_.T)
        self.pc_data.columns = pc_name
        self.pc_data.set_index(DecompositionWorkflowBase.X.columns, inplace=True)
        print(self.pc_data)

    def _get_explained_variance_ratio(self) -> None:
        """Get explained variance ratio."""
        print("-----* Explained Variance Ratio *-----")
        print(self.model.explained_variance_ratio_)

    @staticmethod
    def _biplot(reduced_data: pd.DataFrame, pc_data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Draw bi-plot."""
        print("-----* Compositional Bi-plot *-----")
        biplot(reduced_data, pc_data, algorithm_name)
        save_fig(f"Compositional Bi-plot - {algorithm_name}", local_path, mlflow_path)
        save_data(reduced_data, "Compositional Bi-plot - Reduced Data", local_path, mlflow_path)
        save_data(pc_data, "Compositional Bi-plot - PC Data", local_path, mlflow_path)

    @staticmethod
    def _triplot(reduced_data: pd.DataFrame, pc_data: pd.DataFrame, algorithm_name: str, local_path: str, mlflow_path: str) -> None:
        """Draw tri-plot."""
        print("-----* Compositional Tri-plot *-----")
        triplot(reduced_data, pc_data, algorithm_name)
        save_fig(f"Compositional Tri-plot - {algorithm_name}", local_path, mlflow_path)
        save_data(reduced_data, "Compositional Tri-plot - Reduced Data", local_path, mlflow_path)
        save_data(pc_data, "Compositional Tri-plot - PC Data", local_path, mlflow_path)

    def special_components(self, **kwargs: Union[Dict, np.ndarray, int]) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""
        self._reduced_data2pd(kwargs["reduced_data"], kwargs["components_num"])
        self._get_principal_components()
        self._get_explained_variance_ratio()

        GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH")
        # Draw graphs when the number of principal components > 3
        if kwargs["components_num"] > 3:
            # choose two of dimensions to draw
            two_dimen_axis_index, two_dimen_pc_data = self.choose_dimension_data(self.pc_data, 2)
            two_dimen_reduced_data = self.X_reduced.iloc[:, two_dimen_axis_index]
            self._biplot(
                reduced_data=two_dimen_reduced_data,
                pc_data=two_dimen_pc_data,
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )

            # choose three of dimensions to draw
            three_dimen_axis_index, three_dimen_pc_data = self.choose_dimension_data(self.pc_data, 3)
            three_dimen_reduced_data = self.X_reduced.iloc[:, three_dimen_axis_index]
            self._triplot(
                reduced_data=three_dimen_reduced_data,
                pc_data=three_dimen_pc_data,
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
        elif kwargs["components_num"] == 3:
            # choose two of dimensions to draw
            two_dimen_axis_index, two_dimen_pc_data = self.choose_dimension_data(self.pc_data, 2)
            two_dimen_reduced_data = self.X_reduced.iloc[:, two_dimen_axis_index]
            self._biplot(
                reduced_data=two_dimen_reduced_data,
                pc_data=two_dimen_pc_data,
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
            # no need to choose
            self._triplot(
                reduced_data=self.X_reduced,
                pc_data=self.pc_data,
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
        elif kwargs["components_num"] == 2:
            self._biplot(
                reduced_data=self.X_reduced,
                pc_data=self.pc_data,
                algorithm_name=self.naming,
                local_path=GEOPI_OUTPUT_ARTIFACTS_IMAGE_MODEL_OUTPUT_PATH,
                mlflow_path=MLFLOW_ARTIFACT_IMAGE_MODEL_OUTPUT_PATH,
            )
        else:
            pass


class TSNEDecomposition(DecompositionWorkflowBase):
    """The automation workflow of using T-SNE algorithm to make insightful products."""

    name = "T-SNE"
    special_function = []

    def __init__(
        self,
        n_components: int = 2,
        *,
        perplexity: float = 30.0,
        early_exaggeration: float = 12.0,
        learning_rate: Union[float, str] = "auto",
        n_iter: int = 1000,
        n_iter_without_progress: int = 300,
        min_grad_norm: float = 1e-7,
        metric: str = "euclidean",
        metric_params: Optional[Dict] = None,
        init: str = "pca",
        verbose: int = 0,
        random_state: Optional[int] = None,
        method: str = "exact",
        angle: float = 0.5,
        n_jobs: int = None,
        square_distances: Union[bool, str] = "deprecated",
    ) -> None:
        """
        Parameters
        ----------
        n_components : int, default=2
            Dimension of the embedded space.

        perplexity : float, default=30.0
            The perplexity is related to the number of nearest neighbors that
            is used in other manifold learning algorithms. Larger datasets
            usually require a larger perplexity. Consider selecting a value
            between 5 and 50. Different values can result in significantly
            different results. The perplexity must be less than the number
            of samples.

        early_exaggeration : float, default=12.0
            Controls how tight natural clusters in the original space are in
            the embedded space and how much space will be between them. For
            larger values, the space between natural clusters will be larger
            in the embedded space. Again, the choice of this parameter is not
            very critical. If the cost function increases during initial
            optimization, the early exaggeration factor or the learning rate
            might be too high.

        learning_rate : float or "auto", default="auto"
            The learning rate for t-SNE is usually in the range [10.0, 1000.0]. If
            the learning rate is too high, the data may look like a 'ball' with any
            point approximately equidistant from its nearest neighbours. If the
            learning rate is too low, most points may look compressed in a dense
            cloud with few outliers. If the cost function gets stuck in a bad local
            minimum increasing the learning rate may help.
            Note that many other t-SNE implementations (bhtsne, FIt-SNE, openTSNE,
            etc.) use a definition of learning_rate that is 4 times smaller than
            ours. So our learning_rate=200 corresponds to learning_rate=800 in
            those other implementations. The 'auto' option sets the learning_rate
            to `max(N / early_exaggeration / 4, 50)` where N is the sample size,
            following [4] and [5].

            .. versionchanged:: 1.2
            The default value changed to `"auto"`.

        n_iter : int, default=1000
            Maximum number of iterations for the optimization. Should be at
            least 250.

        n_iter_without_progress : int, default=300
            Maximum number of iterations without progress before we abort the
            optimization, used after 250 initial iterations with early
            exaggeration. Note that progress is only checked every 50 iterations so
            this value is rounded to the next multiple of 50.

            .. versionadded:: 0.17
            parameter *n_iter_without_progress* to control stopping criteria.

        min_grad_norm : float, default=1e-7
            If the gradient norm is below this threshold, the optimization will
            be stopped.

        metric : str or callable, default='euclidean'
            The metric to use when calculating distance between instances in a
            feature array. If metric is a string, it must be one of the options
            allowed by scipy.spatial.distance.pdist for its metric parameter, or
            a metric listed in pairwise.PAIRWISE_DISTANCE_FUNCTIONS.
            If metric is "precomputed", X is assumed to be a distance matrix.
            Alternatively, if metric is a callable function, it is called on each
            pair of instances (rows) and the resulting value recorded. The callable
            should take two arrays from X as input and return a value indicating
            the distance between them. The default is "euclidean" which is
            interpreted as squared euclidean distance.

        metric_params : dict, default=None
            Additional keyword arguments for the metric function.

            .. versionadded:: 1.1

        init : {"random", "pca"} or ndarray of shape (n_samples, n_components), \
                default="pca"
            Initialization of embedding.
            PCA initialization cannot be used with precomputed distances and is
            usually more globally stable than random initialization.

            .. versionchanged:: 1.2
            The default value changed to `"pca"`.

        verbose : int, default=0
            Verbosity level.

        random_state : int, RandomState instance or None, default=None
            Determines the random number generator. Pass an int for reproducible
            results across multiple function calls. Note that different
            initializations might result in different local minima of the cost
            function. See :term:`Glossary <random_state>`.

        method : {'barnes_hut', 'exact'}, default='barnes_hut'
            By default the gradient calculation algorithm uses Barnes-Hut
            approximation running in O(NlogN) time. method='exact'
            will run on the slower, but exact, algorithm in O(N^2) time. The
            exact algorithm should be used when nearest-neighbor errors need
            to be better than 3%. However, the exact method cannot scale to
            millions of examples.

            .. versionadded:: 0.17
            Approximate optimization *method* via the Barnes-Hut.

        angle : float, default=0.5
            Only used if method='barnes_hut'
            This is the trade-off between speed and accuracy for Barnes-Hut T-SNE.
            'angle' is the angular size (referred to as theta in [3]) of a distant
            node as measured from a point. If this size is below 'angle' then it is
            used as a summary node of all points contained within it.
            This method is not very sensitive to changes in this parameter
            in the range of 0.2 - 0.8. Angle less than 0.2 has quickly increasing
            computation time and angle greater 0.8 has quickly increasing error.

        n_jobs : int, default=None
            The number of parallel jobs to run for neighbors search. This parameter
            has no impact when ``metric="precomputed"`` or
            (``metric="euclidean"`` and ``method="exact"``).
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
            ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
            for more details.

            .. versionadded:: 0.22

        square_distances : True, default='deprecated'
            This parameter has no effect since distance values are always squared
            since 1.1.

            .. deprecated:: 1.1
                `square_distances` has no effect from 1.1 and will be removed in
                1.3.

        References
        ----------
        Scikit-learn API: sklearn.manifold.TSNE
        https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
        """
        super().__init__()
        self.n_components = n_components
        self.perplexity = perplexity
        self.early_exaggeration = early_exaggeration
        self.learning_rate = learning_rate
        self.n_iter = n_iter
        self.n_iter_without_progress = n_iter_without_progress
        self.min_grad_norm = min_grad_norm
        self.metric = metric
        self.metric_params = metric_params
        self.init = init
        self.verbose = verbose
        self.method = method
        self.angle = angle
        self.n_jobs = n_jobs
        self.square_distances = square_distances

        if random_state:
            self.random_state = random_state

        # If 'random_state' is None, 'self.random_state' comes from the parent class 'WorkflowBase'
        self.model = TSNE(
            n_components=self.n_components,
            perplexity=self.perplexity,
            early_exaggeration=self.early_exaggeration,
            learning_rate=self.learning_rate,
            n_iter=self.n_iter,
            n_iter_without_progress=self.n_iter_without_progress,
            min_grad_norm=self.min_grad_norm,
            metric=self.metric,
            metric_params=self.metric_params,
            init=self.init,
            verbose=self.verbose,
            random_state=self.random_state,
            method=self.method,
            angle=self.angle,
            n_jobs=self.n_jobs,
            square_distances=self.square_distances,
        )

        self.naming = TSNEDecomposition.name

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = tsne_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""
        pass


class MDSDecomposition(DecompositionWorkflowBase):
    """The automation workflow of using MDS algorithm to make insightful products."""

    name = "MDS"
    special_function = []

    def __init__(
        self,
        n_components: int = 2,
        *,
        metric: bool = True,
        n_init: int = 4,
        max_iter: int = 300,
        verbose: int = 0,
        eps: float = 1e-3,
        n_jobs: Optional[int] = None,
        random_state: Optional[int] = None,
        dissimilarity: str = "euclidean",
        # normalized_stress="warn",
    ) -> None:
        """
        Parameters
        ----------
        n_components : int, default=2
            Number of dimensions in which to immerse the dissimilarities.

        metric : bool, default=True
            If ``True``, perform metric MDS; otherwise, perform nonmetric MDS.
            When ``False`` (i.e. non-metric MDS), dissimilarities with 0 are considered as
            missing values.

        n_init : int, default=4
            Number of times the SMACOF algorithm will be run with different
            initializations. The final results will be the best output of the runs,
            determined by the run with the smallest final stress.

        max_iter : int, default=300
            Maximum number of iterations of the SMACOF algorithm for a single run.

        verbose : int, default=0
            Level of verbosity.

        eps : float, default=1e-3
            Relative tolerance with respect to stress at which to declare
            convergence. The value of `eps` should be tuned separately depending
            on whether or not `normalized_stress` is being used.

        n_jobs : int, default=None
            The number of jobs to use for the computation. If multiple
            initializations are used (``n_init``), each run of the algorithm is
            computed in parallel.

            ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
            ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
            for more details.

        random_state : int, RandomState instance or None, default=None
            Determines the random number generator used to initialize the centers.
            Pass an int for reproducible results across multiple function calls.
            See :term:`Glossary <random_state>`.

        dissimilarity : {'euclidean', 'precomputed'}, default='euclidean'
            Dissimilarity measure to use:

            - 'euclidean':
                Pairwise Euclidean distances between points in the dataset.

            - 'precomputed':
                Pre-computed dissimilarities are passed directly to ``fit`` and
                ``fit_transform``.

        normalized_stress : bool or "auto" default=False
            Whether use and return normed stress value (Stress-1) instead of raw
            stress calculated by default. Only supported in non-metric MDS.

            .. versionadded:: 1.2

        References
        ----------
        Scikit-learn API: sklearn.manifold.MDS
        https://scikit-learn.org/stable/modules/generated/sklearn.manifold.MDS.html
        """
        super().__init__()
        self.n_components = n_components
        self.metric = metric
        self.n_init = n_init
        self.max_iter = max_iter
        self.verbose = verbose
        self.eps = eps
        self.n_jobs = n_jobs
        self.dissimilarity = dissimilarity
        # self.normalized_stress = normalized_stress

        if random_state:
            self.random_state = random_state

        # If 'random_state' is None, 'self.random_state' comes from the parent class 'WorkflowBase'
        self.model = MDS(
            n_components=self.n_components,
            metric=self.metric,
            n_init=self.n_init,
            max_iter=self.max_iter,
            verbose=self.verbose,
            eps=self.eps,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            dissimilarity=self.dissimilarity,
            # normalized_stress=self.normalized_stress,
        )

        self.naming = MDSDecomposition.name

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = mds_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""
        pass



# -*- coding: utf-8 -*-
import random

import numpy as np
import pandas as pd
from rich import print
from scipy.stats import kruskal, wilcoxon


def test_once(df_orig: pd.DataFrame, df_impute: pd.DataFrame, test: str) -> np.ndarray:
    """Do hypothesis testing on each pair-wise column once, non-parametric test.
    Null hypothesis: the distributions of the data set before and after imputing remain the same.

    Parameters
    ----------
    df_orig : pd.DataFrame (n_samples, n_components)
        The original dataset with missing value.

    df_impute : pd.DataFrame (n_samples, n_components)
        The dataset after imputation.

    test : str
        The statistics test method used.

    Returns
    -------
    pvals : np.ndarray
        A numpy array containing the p-values of the tests on each column in the column order
    """
    cols = df_orig.columns
    pvals = np.array([])

    if test == "wilcoxon":
        for c in cols:
            try:
                df_new_orig = df_orig[c].dropna()
                stat, pval = wilcoxon(df_new_orig, df_impute[c])
                pvals = np.append(pvals, pval)
            except Exception:
                pvals = np.append(pvals, 0)

    if test == "kruskal":
        for c in cols:
            df_new_orig = df_orig[c].dropna()
            stat, pval = kruskal(df_new_orig, df_impute[c], nan_policy="omit")
            pvals = np.append(pvals, pval)

    return pvals


def monte_carlo_simulator(
    df_orig: pd.DataFrame,
    df_impute: pd.DataFrame,
    sample_size: int,
    iteration: int,
    test: str,
    confidence: float = 0.05,
) -> None:
    """Check which column rejects hypothesis testing, p value < significance level, to find whether
    the imputation change the distribution of the original data set.

    Parameters
    ----------
    df_orig : pd.DataFrame (n_samples, n_components)
        The original dataset with missing value.

    df_impute : pd.DataFrame (n_samples, n_components)
        The dataset after imputation.

    test : str
        The statistics test method used.

    sample_size : int
        The size of the sample for each iteration.

    iteration : int
        The number of iterations of Monte Carlo simulation.

    confidence : float
        Confidence level, default to be 0.05
    """
    random.seed(2)
    simu_pvals = np.array([0] * df_orig.shape[1])
    for i in range(iteration):
        # monte carlo sampling
        sample_idx = random.sample(range(df_orig.shape[0]), sample_size)
        sample_orig = df_orig.iloc[sample_idx]
        sample_impute = df_impute.iloc[sample_idx]

        # hypothesis testing, non-parametric test
        one_pval = test_once(df_orig=sample_orig, df_impute=sample_impute, test=test)
        simu_pvals = simu_pvals + one_pval

    # average p value
    col_res = simu_pvals / iteration
    # check which column rejects hypothesis testing, 0 < p value < significance level
    rejected_col = df_orig.columns[np.where((col_res < confidence) & (col_res > 0))[0]]

    print("Significance Level: ", confidence)
    print("The number of iterations of Monte Carlo simulation: ", iteration)
    print("The size of the sample for each iteration (half of the whole data set): ", sample_size)
    print("Average p-value: ")
    print("\n".join("{} {}".format(x, y) for x, y in zip(df_orig.columns, col_res)))
    print("Note: 'p-value < 0.05' means imputation method doesn't apply to that column.")
    print("The columns which rejects null hypothesis: ", end="")
    print("None") if not rejected_col.size else print(*list(rejected_col))



# -*- coding: utf-8 -*-
from typing import Dict

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input


def xgboost_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Estimators: The number of trees in the forest.")
    print("Please specify the number of trees in the forest. A good starting range could be between 50 and 500, such as 100.")
    n_estimators = num_input(SECTION[2], "@N Estimators: ")
    print("Learning Rate: It controls the step-size in updating the weights. It shrinks the contribution of each tree by learning_rate. There is a trade-off between learning_rate and n_estimators.")
    print("Please specify the initial learning rate of Xgboost, such as 0.1.")
    learning_rate = float_input(0.01, SECTION[2], "@Learning Rate: ")
    print("Max Depth: The maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit.")
    print("Please specify the maximum depth of a tree. A good starting range could be between 3 and 15, such as 4.")
    max_depth = num_input(SECTION[2], "@Max Depth: ")
    print("Subsample: The fraction of samples to be used for fitting the individual base learners.")
    print("Please specify the fraction of samples to be used for fitting the individual base learners. A good starting range could be between 0.5 and 1.0, such as 0.8.")
    subsample = float_input(1, SECTION[2], "@Subsample: ")
    print("Colsample Bytree: The fraction of features to be used for fitting the individual base learners.")
    print("Please specify the fraction of features to be used for fitting the individual base learners. A good starting range could be between 0.5 and 1.0, such as 1.")
    colsample_bytree = float_input(1, SECTION[2], "@Colsample Bytree: ")
    print("Alpha: L1 regularization term on weights.")
    print("Please specify the L1 regularization term on weights. A good starting range could be between 0 and 1.0, such as 0.")
    alpha = float_input(0, SECTION[2], "@Alpha: ")
    print("Lambda: L2 regularization term on weights.")
    print("Please specify the L2 regularization term on weights. A good starting range could be between 0 and 1.0, such as 1.")
    lambd = float_input(1, SECTION[2], "@Lambda: ")
    hyper_parameters = {
        "n_estimators": n_estimators,
        "learning_rate": learning_rate,
        "max_depth": max_depth,
        "subsample": subsample,
        "colsample_bytree": colsample_bytree,
        "alpha": alpha,
        "lambd": lambd,
    }
    return hyper_parameters


def plot_feature_importance(columns_name: pd.Index, feature_importance: np.ndarray, image_config: Dict) -> None:
    """Plot the Feature Importance.

    Parameters
    ----------
    columns_name : pd.Index
        The input data.

    feature_importance : np.ndarray
        The feature importance value.
    """
    # print the feature importance value orderly
    for feature_name, score in zip(list(columns_name), feature_importance):
        print(feature_name, ":", score)
    plt.figure(figsize=(image_config["width"], image_config["height"]), dpi=image_config["dpi"])
    plt.bar(range(len(columns_name)), feature_importance, tick_label=columns_name)


# def histograms_feature_weights(X: pd.DataFrame, trained_model: object, image_config: dict) -> None:
#     """Plot the Feature Importance, histograms present feature weights for XGBoost predictions.

#     Parameters
#     ----------
#     X: pd.DataFrame (n_samples, n_components)
#         The input data.

#     trained_model : sklearn algorithm model
#         The sklearn algorithm model trained with X_train data.
#     """

#     # columns_name = X.columns
#     plt.rcParams["figure.figsize"] = (image_config["width"], image_config["height"])
#     xgboost.plot_importance(trained_model)
